﻿using System;
using System.Diagnostics;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Helpers;
using SGLF.Graphics;
using SGLF.Graphics.Cameras;
using SGLF.Graphics.Lights;

namespace DeferredRenderer
{
    public class Application : GameWindow
    {
        private Entity entity1;
        private Camera camera;
        private Light[] lights = new Light[48];
        private Vector3[] lightsPosition = new Vector3[48];
        private Renderer renderer;

        public Application()
            : base(640, 480, GraphicsMode.Default, "Deferred Renderer", GameWindowFlags.Default,
                   DisplayDevice.Default, 4, 4, GraphicsContextFlags.Default)
        {
            VSync = VSyncMode.Off;
        }

        protected override void OnLoad(EventArgs e)
        {
            // Setup renderer.
            renderer = new RendererDeferred();
            //renderer = new RendererForward();

            // Setup light.
            Random random = new Random();
            for (int i = 0; i < lights.Length; i++)
            {
                int a = 360 * 180 * i / lights.Length;
                float phi = MathHelper.DegreesToRadians(a % 360);
                float theta = MathHelper.DegreesToRadians(a / 360);

                lights[i] = new Light();
                lights[i].LightType = 1;
                lights[i].Irradiance = ((float) random.NextDouble() * 10) * new Vector3((float) random.NextDouble(), (float) random.NextDouble(), (float) random.NextDouble());
                lights[i].Radius = 6.0f;
                lights[i].Parameters = new Vector3(0.5f, 0, 0);
                lightsPosition[i] = Vector3.UnitY + Vector3Helper.SphericalCoordinates(phi, theta, 3.2f);
                renderer.Add(lights[i]);
            }

            // Setup entities.
            entity1 = new Entity("test", GraphicsManager.Models.Load("moped.mesh"));
            entity1.Transform.Position = new Vector3(0, 0, 0);
            renderer.Add(entity1);

            // Setup camera.
            camera = new PerspectiveCamera(60.0f, 0.1f, 20.0f);
            camera.ClearColor = Color4.Gray;
            camera.Transform.LookAt(6 * Vector3.UnitZ + 1.5f * Vector3.UnitY, 1.5f * Vector3.UnitY, Vector3.UnitY);
            renderer.Add(camera);
        }

        private float elapsedTime = 0;
        private float fpsTime = 0;
        private float fpsAccum = 0;
        private float fps;
        private int frameCount;

        protected override void OnRenderFrame(FrameEventArgs e)
        {
            elapsedTime += (float) e.Time;

            for (int i = 0; i < lights.Length; i++)
            {
                lights[i].Transform.Position = Vector3.Transform(lightsPosition[i], Quaternion.FromAxisAngle(Vector3.UnitY, elapsedTime));
                lights[i].Transform.LookAt(Vector3.Zero, Vector3.UnitY);
            }

            entity1.Transform.Rotation = Quaternion.FromAxisAngle(Vector3.UnitY, -elapsedTime * 0.1f);

            renderer.Width = Width;
            renderer.Height = Height;
            renderer.Render();

            ErrorCode error = GL.GetError();
            if (error != ErrorCode.NoError)
                Debug.WriteLine(error);

            SwapBuffers();

            fpsTime += (float) e.Time;

            if (fpsTime > 1.0f)
            {
                fps = fpsAccum / frameCount;
                fpsAccum = 0;
                frameCount = 0;
                fpsTime -= 1.0f;
                Title = "FPS: " + fps;
            }
            else
            {
                fpsAccum += (float) RenderFrequency;
                frameCount++;
            }
        }

        // Events
        protected override void OnUnload(EventArgs e)
        {
            GraphicsManager.Dispose();
        }

        protected override void OnUpdateFrame(FrameEventArgs e)
        {
        }

        private static void Main(string[] args)
        {
            Debug.Listeners.Add(new TextWriterTraceListener(Console.Out));
            Application program = new Application();
            program.Run();
        }
    }
}