﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

namespace SGLF.Core.Resources
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// A generic <see cref="Resource"/> manager, able to load and release resources. Everytime a
    /// <see cref="Resource"/> is loaded, a reference is stored in an available resources list. Therefore,
    /// each subsequent request of that very resource returns the already loaded instance, instead of loading
    /// it anew.
    ///
    /// When the <see cref="Resource"/> is no longer needed, it'll be disposed automatically. This condition
    /// is asserted through the use of reference counting: each resource has an associated counter that
    /// keeps track of how many objects are using it. At each <c>Load</c> method call, the counter will be
    /// incremented by 1. Likewise, at each <c>Release</c>, this counter is decremented by 1. Upon reaching
    /// a reference count of 0, the <see cref="Resource"/> will be disposed.
    /// </summary>
    /// <typeparam name="T">A <see cref="Resource"/> type.</typeparam>
    public abstract class ResourceManager<T>
        where T : Resource
    {
        // ==================================================================================================
        // Structures
        // ==================================================================================================

        /// <summary>
        /// A simple triplet class holding both the resource and it's reference count.
        /// </summary>
        private class ResourceTracker
        {
            public string Name
            {
                get;
                set;
            }

            public T Resource
            {
                get;
                set;
            }

            public int RefCount
            {
                get;
                set;
            }
        }

        // ==================================================================================================
        // Members
        // ==================================================================================================

        private Dictionary<string, ResourceTracker> ResourcesByName;
        private Dictionary<T, ResourceTracker> ResourcesByRef;

        // ==================================================================================================
        // Properties
        // ==================================================================================================

        /// <summary>
        /// The base path added to each request in order to locate the resource.
        /// </summary>
        public string BasePath
        {
            get;
            set;
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="ResourceManager{T}"/> class.
        /// </summary>
        public ResourceManager()
        {
            ResourcesByName = new Dictionary<string, ResourceTracker>();
            ResourcesByRef = new Dictionary<T, ResourceTracker>();
        }

        /// <summary>
        /// Releases all resources.
        /// </summary>
        public void Dispose()
        {
            foreach (var resourceRefCount in ResourcesByName.Values)
                resourceRefCount.Resource.Dispose(true);

            ResourcesByName.Clear();
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        /// <summary>
        /// Registers an existing <see cref="Resource"/> instance to the manager.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="resource">The resource.</param>
        public void Register(string name, T resource)
        {
            // Check resource.
            if (resource == null)
                throw new ArgumentNullException("resource");
            if (resource.Disposed)
                throw new ObjectDisposedException("resource");

            // Check if it already exists.
            if (ResourcesByName.ContainsKey(name))
                throw new ArgumentException("name");

            // Initialize resource tracker.
            ResourceTracker resourceTracking = new ResourceTracker();
            resourceTracking.Name = name;
            resourceTracking.Resource = resource;
            resourceTracking.RefCount = 1;

            // Register.
            ResourcesByName.Add(name, resourceTracking);
            ResourcesByRef.Add(resource, resourceTracking);
        }

        /// <summary>
        /// Tries to load a <see cref="Resource"/> with the given name.
        /// </summary>
        /// <param name="name">The resource name.</param>
        /// <returns>The <see cref="Resource"/> instance. <c>null</c> if it wasn't possible to load.</returns>
        public T Load(string name)
        {
            string fileName = Path.Combine(BasePath, name);
            ResourceTracker resourceTracking = null;

            if (!ResourcesByName.TryGetValue(name, out resourceTracking))
            {
                try
                {
                    T resource = LoadFromFile(fileName);
                    Register(name, resource);

                    return resource;
                }
                catch (Exception ex)
                {
                    Debug.WriteLine("Unable to load " + typeof(T).Name);
                    Debug.WriteLine("- File: \"" + fileName + "\"");
                    Debug.WriteLine("- Exception: \"" + ex.GetType().Name + "\"");
                    Debug.WriteLine("- Message: \"" + ex.Message + "\"");
                    throw ex;
                }
            }
            else
            {
                return resourceTracking.Resource;
            }
        }

        /// <summary>
        /// Releases the specified resource.
        /// </summary>
        /// <param name="resource">The resource.</param>
        public void Release(T resource)
        {
            ResourceTracker resourceTracking = null;

            if (ResourcesByRef.TryGetValue(resource, out resourceTracking))
            {
                if (--resourceTracking.RefCount == 0)
                {
                    ResourcesByName.Remove(resourceTracking.Name);
                    ResourcesByRef.Remove(resourceTracking.Resource);
                    resource.Dispose();
                }
            }
            else
                throw new ArgumentException("resource");
        }

        /// <summary>
        /// Abstract method to be implemented by the specific manager. Loads a new instance of the
        /// <see cref="Resource"/> object located at the given path.
        /// </summary>
        /// <param name="filePath">Path to the file.</param>
        /// <returns>The <see cref="Resource"/> instance.</returns>
        protected abstract T LoadFromFile(string filePath);
    }
}