﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.InteropServices;
using OpenTK.Graphics.OpenGL;

namespace SGLF.Graphics.Buffers
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Represents an OpenGL buffer object.
    /// </summary>
    public abstract class AbstractBuffer : GraphicsNamedResource
    {
        // ==================================================================================================
        // Properties
        // ==================================================================================================

        /// <summary>
        /// The buffer target.
        /// </summary>
        public BufferTarget Target
        {
            get;
            private set;
        }

        /// <summary>
        /// The buffer usage hint.
        /// </summary>
        public BufferUsageHint UsageHint
        {
            get;
            private set;
        }

        /// <summary>
        /// The size (in bytes) of the buffer.
        /// </summary>
        public int SizeInBytes
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets a value indicating whether the buffer is currently mapped.
        /// </summary>
        public bool Mapped
        {
            get
            {
                return Mapping != IntPtr.Zero;
            }
        }

        /// <summary>
        /// Gets the pointer to the current buffer mapping.
        /// </summary>
        public IntPtr Mapping
        {
            get;
            private set;
        }

        /// <summary>
        /// The buffer access permissions.
        /// </summary>
        public BufferAccess Access
        {
            get
            {
                int access;
                GL.Ext.GetNamedBufferParameter(Handle, (ExtDirectStateAccess)
                                               BufferParameterName.BufferAccess,
                                               out access);
                return (BufferAccess) access;
            }
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="AbstractBuffer"/> class.
        /// </summary>
        /// <param name="target">The buffer target.</param>
        /// <param name="usageHint">The buffer usage hint.</param>
        /// <param name="sizeInBytes">The buffer size (in bytes).</param>
        /// <exception cref="System.ArgumentOutOfRangeException">If sizeInBytes is less or equal
        /// than 0.</exception>
        public AbstractBuffer(BufferTarget target, BufferUsageHint usageHint, int sizeInBytes)
        {
            if (sizeInBytes <= 0)
                throw new ArgumentOutOfRangeException("sizeInBytes");

            Handle = GL.GenBuffer();

            Target = target;
            UsageHint = usageHint;
            Resize(sizeInBytes);
        }

        /// <summary>
        /// Disposes the instance's resources.
        /// </summary>
        /// <param name="disposeManaged"><c>true</c> to release managed resources.</param>
        /// <param name="disposeUnmanaged"><c>true</c> to release unmanaged resources.</param>
        protected override void Dispose(bool disposeManaged, bool disposeUnmanaged)
        {
            if (disposeUnmanaged)
                GL.DeleteBuffer(Handle);

            base.Dispose(disposeManaged, disposeUnmanaged);
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        /// <summary>
        /// Resizes the buffer's storage.
        /// </summary>
        /// <param name="newSizeInBytes">The new size (in bytes).</param>
        protected void Resize(int sizeInBytes)
        {
            if (sizeInBytes <= 0)
                throw new ArgumentOutOfRangeException("sizeInBytes");

            SizeInBytes = sizeInBytes;
            GL.Ext.NamedBufferData(Handle, (IntPtr) SizeInBytes, IntPtr.Zero, (ExtDirectStateAccess) UsageHint);
        }

        /// <summary>
        /// Gets the contents of the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data read will begin.</param>
        /// <param name="length">Specifies the size (in bytes) of the data store region being
        /// read.</param>
        /// <param name="data">Specifies the pointer to the memory block that will store the buffer
        /// contents.</param>
        protected void GetRawData(int offset, int length, IntPtr data)
        {
            // Check data.
            if (data == IntPtr.Zero)
                throw new ArgumentNullException("data");

            // Check offset & length.
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("length");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("length");

            // Get data.
            GL.Ext.GetNamedBufferSubData(Handle, (IntPtr) offset, (IntPtr) length, data);
        }

        /// <summary>
        /// Gets the contents of the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data read will begin.</param>
        /// <param name="data">Specifies the array that will store the buffer contents.</param>
        protected void GetRawData(int offset, Array data)
        {
            // Check data.
            if (data == null)
                throw new ArgumentNullException("data");

            // Get data sizes.
            Type type = data.GetType().GetElementType();
            int stride = Marshal.SizeOf(type);
            int length = stride * data.Length;

            // Check offset & length.
            if (data == null)
                throw new ArgumentNullException("data");
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");

            // Get data.
            GCHandle dataHandle = GCHandle.Alloc(data, GCHandleType.Pinned);
            IntPtr dataPointer = dataHandle.AddrOfPinnedObject();
            GL.Ext.GetNamedBufferSubData(Handle, (IntPtr) offset, (IntPtr) length, dataPointer);
            dataHandle.Free();
        }

        /// <summary>
        /// Gets the contents of the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data read will begin.</param>
        /// <param name="data">Specifies the array that will store the buffer contents.</param>
        protected void GetRawData<T>(int offset, T[] data)
            where T : struct
        {
            // Check data.
            if (data == null)
                throw new ArgumentNullException("data");

            // Get data sizes.
            int stride = Marshal.SizeOf(default(T));
            int length = stride * data.Length;

            // Check offset & length.
            if (data == null)
                throw new ArgumentNullException("data");
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");

            // Get data.
            GL.Ext.GetNamedBufferSubData(Handle, (IntPtr) offset, (IntPtr) length, data);
        }

        /// <summary>
        /// Streams from nested subset of data the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data read will begin.</param>
        /// <param name="stride">Specified the stride between read operations.</param>
        /// <param name="data">Specifies the array that will store the buffer contents.</param>
        protected void GetRawStreamData<T>(int offset, int stride, T[] data)
            where T : struct
        {
            // TODO: Implement :(
            throw new NotImplementedException();
        }

        /// <summary>
        /// Sets the supplied data to the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data replacement will begin.</param>
        /// <param name="length">Specifies the size (in bytes) of the data store region being
        /// written.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the
        /// buffer.</param>
        protected void SetRawData(int offset, int length, IntPtr data)
        {
            // Check data.
            if (data == IntPtr.Zero)
                throw new ArgumentNullException("data");

            // Check offset & length.
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("length");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("length");

            // Set data.
            GL.Ext.NamedBufferSubData(Handle, (IntPtr) offset, (IntPtr) length, data);
        }

        /// <summary>
        /// Sets the supplied data to the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data replacement will begin.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the
        /// buffer.</param>
        protected void SetRawData(int offset, Array data)
        {
            // Check data.
            if (data == null)
                throw new ArgumentNullException("data");

            // Get data sizes.
            Type type = data.GetType().GetElementType();
            int stride = Marshal.SizeOf(type);
            int length = stride * data.Length;

            // Check offset & length.
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");

            // Set data.
            GCHandle dataHandle = GCHandle.Alloc(data, GCHandleType.Pinned);
            IntPtr dataPointer = dataHandle.AddrOfPinnedObject();
            GL.Ext.NamedBufferSubData(Handle, (IntPtr) offset, (IntPtr) length, dataPointer);
            dataHandle.Free();
        }

        /// <summary>
        /// Sets the supplied data to the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data replacement will begin.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the
        /// buffer.</param>
        protected void SetRawData<T>(int offset, T[] data)
            where T : struct
        {
            // Check data.
            if (data == null)
                throw new ArgumentNullException("data");

            // Get data sizes.
            int stride = Marshal.SizeOf(default(T));
            int length = stride * data.Length;

            // Check offset & length.
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");

            // Set data.
            GL.Ext.NamedBufferSubData(Handle, (IntPtr) offset, (IntPtr) length, data);
        }

        /// <summary>
        /// Streams a nested subset of data the buffer.
        /// </summary>
        /// <param name="offset">Specifies the offset (in bytes) into the buffer object's data
        /// store where data replacement will begin.</param>
        /// <param name="stride">Specified the stride between copy operations.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the
        /// buffer.</param>
        protected void SetRawStreamData<T>(int offset, int stride, T[] data)
            where T : struct
        {
            // Check data.
            if (data == null)
                throw new ArgumentNullException("data");

            // Get data sizes.
            int count = data.Length;
            int dataStride = Marshal.SizeOf(default(T));
            int dataLength = dataStride * count;
            int length = System.Math.Min(stride * (count - 1), dataStride);

            // Check offset & length.
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (dataStride > stride)
                throw new ArgumentOutOfRangeException("stride");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("data");

            // Map buffer and pin data buffer.
            IntPtr mapping = Map(offset, dataLength, BufferAccessMask.MapWriteBit);

            GCHandle dataHandle = GCHandle.Alloc(data, GCHandleType.Pinned);
            IntPtr dataPointer = dataHandle.AddrOfPinnedObject();

            // Copy byte to byte.
            unsafe
            {
                byte* sourcePtr = (byte*) dataPointer;
                byte* destPtr = (byte*) mapping;

                for (int i = 0; i < count; i++, sourcePtr += dataStride, destPtr += stride)
                {
                    for (int j = 0; j < dataStride; j++)
                    {
                        destPtr[j] = sourcePtr[j];
                    }
                }
            }

            // Done.
            dataHandle.Free();
            Unmap();
        }

        /// <summary>
        /// Maps a subset of the buffer contents into the client's address space.
        /// </summary>
        /// <param name="offset">Specifies a the starting offset within the buffer of the range to
        /// be mapped.</param>
        /// <param name="length">Specifies a length of the range to be mapped.</param>
        /// <param name="access">Specifies a combination of access flags indicating the desired
        /// access to the range.</param>
        /// <returns>Pointer to the beggining of the mapped range.</returns>
        public IntPtr Map(int offset, int length, BufferAccessMask access)
        {
            // Check offset & length.
            if (offset < 0 || offset >= SizeInBytes)
                throw new ArgumentOutOfRangeException("offset");
            if (length > SizeInBytes)
                throw new ArgumentOutOfRangeException("length");
            if ((offset + length) > SizeInBytes)
                throw new ArgumentOutOfRangeException("length");

            // Check mapping status.
            if (Mapped)
                throw new InvalidOperationException("Buffer is already mapped.");

            // Map buffer.
            Mapping = GL.Ext.MapNamedBufferRange(Handle, (IntPtr) offset, (IntPtr) length, access);

            if (Mapping == IntPtr.Zero)
                throw new InvalidOperationException("Unable to map buffer.");

            return Mapping;
        }

        /// <summary>
        /// Unmaps the buffer contents from the client's address space.
        /// </summary>
        public void Unmap()
        {
            // Check mapping status.
            if (!Mapped)
                throw new InvalidOperationException("Buffer is not mapped.");

            // Unmap buffer.
            GL.Ext.UnmapNamedBuffer(Handle);
            Mapping = IntPtr.Zero;
        }
    }
}