﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.InteropServices;
using OpenTK.Graphics.OpenGL;

namespace SGLF.Graphics.Buffers
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    public sealed class UniformBuffer<T> : AbstractBuffer
        where T : struct
    {
        // ==================================================================================================
        // Properties
        // ==================================================================================================

        public int Count
        {
            get;
            private set;
        }

        public int Stride
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Constructor
        // ==================================================================================================

        public UniformBuffer(int count, BufferUsageHint usageHint = BufferUsageHint.StreamDraw)
            : this(count, Marshal.SizeOf(typeof(T)), usageHint)
        {
        }

        public UniformBuffer(int count, int stride, BufferUsageHint usageHint = BufferUsageHint.StreamDraw)
            : base(BufferTarget.UniformBuffer, usageHint, count * stride)
        {
            if (count <= 0)
                throw new ArgumentOutOfRangeException("count");

            Count = count;
            Stride = stride;
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        public new void Resize(int count)
        {
            if (count <= 0)
                throw new ArgumentOutOfRangeException("count");

            base.Resize(Stride * count);
        }

        /// <summary>
        /// Gets the contents of the buffer.
        /// </summary>
        /// <param name="index">Specifies the first index where the data read will begin.</param>
        /// <param name="data">Specifies the array that will store the buffer contents.</param>
        public void GetData(int index, T[] data)
        {
            // Check data / index
            if (data == null)
                throw new ArgumentNullException("data");
            if (index < 0 || index >= Count)
                throw new ArgumentException("index");

            // Get data.
            GetRawData(Stride * index, data);
        }

        /// <summary>
        /// Sets the supplied data to the buffer.
        /// </summary>
        /// <param name="index">Specifies the first index where the data replacement will begin.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the buffer.</param>
        public void SetData(int index, T[] data)
        {
            // Check data / index
            if (data == null)
                throw new ArgumentNullException("data");
            if (index < 0 || index >= Count)
                throw new ArgumentException("index");

            // Set data.
            SetRawData(Stride * index, data);
        }
    }
}