﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using OpenTK.Graphics.OpenGL;
using SGLF.Graphics.Vertices;

namespace SGLF.Graphics.Buffers
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    ///
    /// </summary>
    public sealed class VertexBuffer : AbstractBuffer
    {
        // ==================================================================================================
        // Properties
        // ==================================================================================================

        /// <summary>
        ///
        /// </summary>
        public int VertexCount
        {
            get;
            private set;
        }

        /// <summary>
        ///
        /// </summary>
        public VertexFormat VertexFormat
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Constructor
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="VertexBuffer"/> class.
        /// </summary>
        /// <param name="vertexFormat">The vertex format.</param>
        /// <param name="vertexCount">The vertex count.</param>
        /// <param name="usageHint">The usage hint.</param>
        public VertexBuffer(VertexFormat vertexFormat, int vertexCount, BufferUsageHint usageHint)
            : base(BufferTarget.ArrayBuffer, usageHint, vertexFormat.Stride * vertexCount)
        {
            if (vertexCount <= 0)
                throw new ArgumentOutOfRangeException("vertexCount");

            VertexFormat = vertexFormat;
            VertexCount = vertexCount;
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        /// <summary>
        /// Gets the contents of the buffer.
        /// </summary>
        /// <param name="index">Specifies the first vertex index where the data read will begin.</param>
        /// <param name="data">Specifies the array that will store the buffer contents.</param>
        public void GetData(int index, Array data)
        {
            // Check data / index
            if (data == null)
                throw new ArgumentNullException("data");
            if (index < 0 || index >= VertexCount)
                throw new ArgumentException("index");

            // Check vertex format.
            Type type = data.GetType().GetElementType();

            if (VertexFormat != VertexFormat.FromType(type))
                throw new ArgumentException("Vertex formats mismatch.");

            // Get data.
            GetRawData(VertexFormat.Stride * index, data);
        }

        /// <summary>
        /// Gets the contents of the buffer.
        /// </summary>
        /// <param name="index">Specifies the first vertex index where the data read will begin.</param>
        /// <param name="data">Specifies the array that will store the buffer contents.</param>
        public void GetData<T>(int index, T[] data)
            where T : struct, IVertex
        {
            // Check data / index
            if (data == null)
                throw new ArgumentNullException("data");
            if (index < 0 || index >= VertexCount)
                throw new ArgumentException("index");

            // Check vertex format.
            if (VertexFormat != VertexFormat.FromType<T>())
                throw new ArgumentException("Vertex formats mismatch.");

            // Get data.
            GetRawData<T>(VertexFormat.Stride * index, data);
        }

        /// <summary>
        /// Sets the supplied data to the buffer.
        /// </summary>
        /// <param name="index">Specifies the first vertex index where the data replacement will begin.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the buffer.</param>
        public void SetData(int index, Array data)
        {
            // Check data / index
            if (data == null)
                throw new ArgumentNullException("data");
            if (index < 0 || index >= VertexCount)
                throw new ArgumentException("index");

            // Check vertex format.
            Type type = data.GetType().GetElementType();

            if (VertexFormat != VertexFormat.FromType(type))
                throw new ArgumentException("Vertex formats mismatch.");

            // Get data.
            SetRawData(VertexFormat.Stride * index, data);
        }

        /// <summary>
        /// Sets the supplied data to the buffer.
        /// </summary>
        /// <param name="index">Specifies the first vertex index where the data replacement will begin.</param>
        /// <param name="data">Specifies the array to the data that will be copied to the buffer.</param>
        public void SetData<T>(int index, T[] data)
            where T : struct, IVertex
        {
            // Check data / index
            if (data == null)
                throw new ArgumentNullException("data");
            if (index < 0 || index >= VertexCount)
                throw new ArgumentException("index");

            // Check vertex format.
            if (VertexFormat != VertexFormat.FromType<T>())
                throw new ArgumentException("Vertex formats mismatch.");

            // Set data.
            SetRawData<T>(VertexFormat.Stride * index, data);
        }
    }
}