﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Threading;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Helpers;
using SGLF.Graphics.Buffers;
using SGLF.Graphics.Shaders;
using SGLF.Graphics.Textures;
using SGLF.Graphics.Vertices;

namespace SGLF.Graphics
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Static Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Represents the current state of the graphics device and binded resources. Warning: we make
    /// the assumption we're using a single graphics context.
    /// </summary>
    public static class GraphicsState
    {
        // ==================================================================================================
        // Members
        // ==================================================================================================

        private static Dictionary<BufferTarget, AbstractBuffer> Buffer;

        private static Dictionary<FramebufferTarget, Framebuffer> Framebuffer;

        private static Program Program;

        private static Dictionary<TextureUnit, TextureObject> Texture;

        private static VertexArrayObject VertexArray;

        // ==================================================================================================
        // Properties
        // ==================================================================================================

        /// <summary>
        /// The <see cref="GraphicsContext"/>.
        /// </summary>
        public static GraphicsContext Context
        {
            get;
            private set;
        }

        /// <summary>
        /// The <see cref="Thread"/> in which the <see cref="GraphicsContext"/> resides.
        /// </summary>
        public static Thread ContextThread
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Static Constructor
        // ==================================================================================================

        /// <summary>
        /// Initializes the <see cref="GraphicsState"/> class.
        /// </summary>
        static GraphicsState()
        {
            // Initialize context.
            Context = (GraphicsContext) GraphicsContext.CurrentContext;
            ContextThread = Thread.CurrentThread;

            // Initialize default state.
            GL.Enable(EnableCap.DepthTest);
            GL.DepthFunc(DepthFunction.Less);
            GL.ClearDepth(1.0f);

            // Initialize state.
            Buffer = EnumHelper.CreateDictionary<BufferTarget, AbstractBuffer>();
            Texture = EnumHelper.CreateDictionary<TextureUnit, TextureObject>();
            Framebuffer = EnumHelper.CreateDictionary<FramebufferTarget, Framebuffer>();
        }

        // ==================================================================================================
        // Static Methods
        // ==================================================================================================

        /// <summary>
        /// Binds a <see cref="AbstractBuffer"/>.
        /// </summary>
        /// <param name="buffer">The buffer.</param>
        public static void BindBuffer(AbstractBuffer buffer)
        {
            if (buffer == null)
                throw new ArgumentNullException("buffer");
            if (buffer.Disposed)
                throw new ObjectDisposedException("buffer");
            if (buffer == Buffer[buffer.Target])
                return;

            GL.BindBuffer(buffer.Target, buffer.Handle);
            Buffer[buffer.Target] = buffer;
        }

        /// <summary>
        /// Binds a <see cref="Framebuffer"/>.
        /// </summary>
        /// <param name="framebuffer">The framebuffer.</param>
        public static void BindFramebuffer(Framebuffer framebuffer)
        {
            if (framebuffer == null)
                throw new ArgumentNullException("framebuffer");
            if (framebuffer.Disposed)
                throw new ObjectDisposedException("framebuffer");
            if (framebuffer == Framebuffer[framebuffer.Target])
                return;
            if (framebuffer.Status != FramebufferErrorCode.FramebufferComplete)
                throw new InvalidOperationException(framebuffer.Status.ToString());

            GL.BindFramebuffer(framebuffer.Target, framebuffer.Handle);
            Framebuffer[framebuffer.Target] = framebuffer;
        }

        /// <summary>
        /// Binds a <see cref="Program"/>.
        /// </summary>
        /// <param name="program">The shader program.</param>
        public static void BindProgram(Program program)
        {
            if (program == null)
                throw new ArgumentNullException("program");
            if (program.Disposed)
                throw new ObjectDisposedException("program");
            if (program == Program)
                return;

            GL.UseProgram(program.Handle);
            Program = program;
        }

        /// <summary>
        /// Binds a <see cref="TextureObject"/>.
        /// </summary>
        /// <param name="unit">The texture unit.</param>
        /// <param name="texture">The texture.</param>
        public static void BindTexture(TextureUnit unit, TextureObject texture)
        {
            if (texture == null)
                throw new ArgumentNullException("texture");
            if (texture.Disposed)
                throw new ObjectDisposedException("texture");
            if (texture == Texture[unit])
                return;

            GL.ActiveTexture(unit);
            GL.BindTexture(texture.Target, texture.Handle);
            Texture[unit] = texture;
        }

        /// <summary>
        /// Binds a <see cref="VertexArrayObject"/>.
        /// </summary>
        /// <param name="vertexArray">The Vertex Array.</param>
        public static void BindVertexArray(VertexArrayObject vertexArray)
        {
            if (vertexArray == null)
                throw new ArgumentNullException("vertexArray");
            if (vertexArray.Disposed)
                throw new ObjectDisposedException("vertexArray");
            if (vertexArray == VertexArray)
                return;

            GL.BindVertexArray(vertexArray.Handle);
            VertexArray = vertexArray;
        }

        /// <summary>
        /// Unbinds the specified <see cref="BufferTarget"/>.
        /// </summary>
        /// <param name="target">The buffer target.</param>
        public static void UnbindBuffer(BufferTarget target)
        {
            if (Buffer[target] == null)
                return;

            GL.BindBuffer(target, 0);
            Buffer[target] = null;
        }

        /// <summary>
        /// Unbinds the current <see cref="Framebuffer"/>.
        /// </summary>
        /// <param name="target">The framebuffer target.</param>
        public static void UnbindFramebuffer(FramebufferTarget target)
        {
            if (Framebuffer[target] == null)
                return;

            GL.BindFramebuffer(target, 0);
            Framebuffer[target] = null;
        }

        /// <summary>
        /// Unbinds the current <see cref="Program"/>.
        /// </summary>
        public static void UnbindProgram()
        {
            if (Program == null)
                return;

            GL.UseProgram(0);
            Program = null;
        }

        /// <summary>
        /// Unbinds the specified <see cref="TextureUnit"/>.
        /// </summary>
        /// <param name="unit">The texture unit.</param>
        public static void UnbindTexture(TextureUnit unit)
        {
            if (Texture[unit] == null)
                return;

            GL.ActiveTexture(unit);
            GL.BindTexture(Texture[unit].Target, 0);
            Texture[unit] = null;
        }

        /// <summary>
        /// Unbinds the current <see cref="VertexArrayObject"/>.
        /// </summary>
        public static void UnbindVertexArray()
        {
            if (VertexArray == null)
                return;

            GL.BindVertexArray(0);
            VertexArray = null;
        }
    }
}