﻿using System;
using System.Collections.Generic;
using System.Xml.Linq;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Extensions;
using SGLF.Core.Helpers;
using SGLF.Core.Resources;
using SGLF.Graphics.Materials;
using SGLF.Graphics.Models;
using SGLF.Graphics.Vertices;

namespace SGLF.Graphics.Managers
{
    public class ModelManager : ResourceManager<Model>
    {
        protected override Model LoadFromFile(string fileName)
        {
            XElement root = XElement.Load(fileName);
            return LoadModel(root);
        }

        private Model LoadModel(XElement root)
        {
            Model model = new Model();
            XElement subMeshes = root.Element("submeshes");

            foreach (var subMesh in subMeshes.Elements("submesh"))
                LoadSubMesh(subMesh, model);

            return model;
        }

        private void LoadSubMesh(XElement subMesh, Model model)
        {
            XElement subMeshFaces = subMesh.Element("faces");
            XElement subMeshGeom = subMesh.Element("geometry");

            bool hasSharedVertices = subMesh.Attribute("usesharedvertices").Value<bool>();
            string materialName = subMesh.Attribute("material").Value;
            string primitiveName = subMesh.Attribute("operationtype").Value;

            if (hasSharedVertices || primitiveName != "triangle_list")
                throw new NotSupportedException();

            int vertexCount = subMeshGeom.Attribute("vertexcount").Value<int>();
            int facesCount = subMeshFaces.Attribute("count").Value<int>();

            VertexFormat vertexFormat = FindSuitableFormat(subMeshGeom);
            PrimitiveType primitiveType = PrimitiveType.Triangles;
            DrawElementsType indexType = DrawElementsType.UnsignedInt;

            Mesh mesh = new Mesh(vertexFormat, vertexCount, primitiveType, 3 * facesCount, indexType, false);
            LoadSubMeshGeometry(subMeshGeom, vertexFormat, vertexCount, mesh);
            LoadSubMeshFaces(subMeshFaces, facesCount, mesh);
            model.Meshes.Add(mesh);

            Material material = GraphicsManager.Materials.Load(materialName + ".material");
            model.Materials.Add(material);
        }

        private void LoadSubMeshFaces(XElement subMeshFaces, int facesCount, Mesh mesh)
        {
            int[] indices = new int[3 * facesCount];
            int index = 0;

            foreach (var face in subMeshFaces.Elements("face"))
            {
                indices[index++] = face.Attribute("v1").Value<int>();
                indices[index++] = face.Attribute("v2").Value<int>();
                indices[index++] = face.Attribute("v3").Value<int>();
            }

            mesh.IndexBuffer.SetData(0, indices);
        }

        private void LoadSubMeshGeometry(XElement subMeshGeom, VertexFormat vertexFormat, int vertexCount, Mesh mesh)
        {
            Array vertices = Array.CreateInstance(vertexFormat.VertexType, vertexCount);
            float radius = 0;

            foreach (var vertexBuffer in subMeshGeom.Elements("vertexbuffer"))
            {
                int index = 0;

                foreach (var vertexData in vertexBuffer.Elements("vertex"))
                {
                    XElement position = vertexData.Element("position");
                    XElement normal = vertexData.Element("normal");
                    XElement tangent = vertexData.Element("tangent");
                    XElement texCoord = vertexData.Element("texcoord");
                    XElement color = vertexData.Element("colour_diffuse");
                    IVertex vertex = (IVertex) vertices.GetValue(index);

                    if (position != null)
                    {
                        vertex.Position = XmlHelper.GetVector3(position);
                        radius = Math.Max(radius, vertex.Position.Length);
                    }

                    if (normal != null)
                        vertex.Normal = XmlHelper.GetVector3(normal);

                    if (tangent != null)
                        vertex.Tangent = XmlHelper.GetVector4(tangent);

                    if (texCoord != null)
                        vertex.TexCoords = XmlHelper.GetVector2(texCoord, "u", "v");

                    if (color != null)
                        vertex.Color = XmlHelper.GetColor(color);

                    vertices.SetValue(vertex, index);
                    index++;
                }
            }

            mesh.BoundingRadius = radius;
            mesh.VertexBuffer.SetData(0, vertices);
        }

        private VertexFormat FindSuitableFormat(XElement subMeshGeom)
        {
            HashSet<VertexSemantics> requiredSemantics = new HashSet<VertexSemantics>();

            foreach (var vertexBuffer in subMeshGeom.Elements("vertexbuffer"))
            {
                if (vertexBuffer.Attribute("positions").Value<bool>())
                    requiredSemantics.Add(VertexSemantics.Position);

                if (vertexBuffer.Attribute("normals").Value<bool>())
                    requiredSemantics.Add(VertexSemantics.Normal);

                if (vertexBuffer.Attribute("tangents").Value<bool>())
                    requiredSemantics.Add(VertexSemantics.Tangent);

                if (vertexBuffer.Attribute("texture_coords").Value<int>(0) > 0)
                    requiredSemantics.Add(VertexSemantics.TexCoord);

                if (vertexBuffer.Attribute("colours_diffuse").Value<bool>())
                    requiredSemantics.Add(VertexSemantics.Color);
            }

            VertexFormat format = VertexFormat.FromSemantics(requiredSemantics);

            if (format == null)
                throw new NotSupportedException("No suitable vertex format found");

            return format;
        }
    }
}