﻿using System;
using System.IO;
using System.Text.RegularExpressions;
using System.Xml.Linq;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Extensions;
using SGLF.Core.Resources;
using SGLF.Graphics.Shaders;

namespace SGLF.Graphics.Managers
{
    public class ProgramManager : ResourceManager<UberProgram>
    {
        private const string RegexBlockComments = @"/\*(.*?)\*/";
        private const string RegexLineComments = @"//(.*?)\r?\n";
        private const string RegexInclude = @"\#include\s+[""<]([^"">]+)*["">]";

        protected override UberProgram LoadFromFile(string fileName)
        {
            XElement root = XElement.Load(fileName);
            return LoadUberProgram(root);
        }

        private UberProgram LoadUberProgram(XElement root)
        {
            UberProgram program = new UberProgram();

            LoadOptions(root.Element("options"), program);
            LoadShaders(root.Element("shaders"), program);

            program.Generate();
            return program;
        }

        private void LoadOptions(XElement subOptions, UberProgram program)
        {
            if (subOptions == null)
                return;

            foreach (var subOption in subOptions.Elements("option"))
                program.SetOption(subOption.Attribute("name").Value());
        }

        private void LoadShaders(XElement subShaders, UberProgram program)
        {
            foreach (var subShader in subShaders.Elements("shader"))
            {
                string fileName = Path.Combine(BasePath, subShader.Attribute("filename").Value());

                ShaderType shaderType = FindShaderType(fileName);
                string sourceCode = LoadShaderSource(fileName);

                program.SetShaderSource(shaderType, sourceCode);
            }
        }

        private string LoadShaderSource(string fileName)
        {
            string sourceCode = File.ReadAllText(fileName);
            sourceCode = Preprocessor(sourceCode);

            return sourceCode;
        }

        private ShaderType FindShaderType(string fileName)
        {
            string extension = Path.GetExtension(fileName);

            switch (extension)
            {
                case ".vert":
                    return ShaderType.VertexShader;

                case ".frag":
                    return ShaderType.FragmentShader;

                case ".geom":
                    return ShaderType.GeometryShader;

                case ".tctrl":
                    return ShaderType.TessControlShader;

                case ".teval":
                    return ShaderType.TessEvaluationShader;

                case ".comp":
                    return ShaderType.ComputeShader;

                default:
                    throw new InvalidOperationException("Unknown shader type.");
            }
        }

        private string Preprocessor(string input)
        {
            PreprocessorRemoveComments(ref input);
            PreprocessorInclude(ref input);
            return input;
        }

        private void PreprocessorRemoveComments(ref string input)
        {
            input = Regex.Replace
            (
                input,
                RegexBlockComments + "|" + RegexLineComments,
                me =>
                {
                    return me.Value.StartsWith("//") ? Environment.NewLine : "";
                },
                RegexOptions.Singleline
            );
        }

        private void PreprocessorInclude(ref string input)
        {
            input = Regex.Replace
            (
                input,
                RegexInclude,
                me =>
                {
                    return LoadShaderSource(Path.Combine(BasePath, me.Groups[1].Value));
                },
                RegexOptions.Multiline
            );
        }
    }
}