﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System.Collections.Generic;
using System.Drawing;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Helpers;
using SGLF.Core.Maths;
using SGLF.Graphics.Cameras;
using SGLF.Graphics.Lights;
using SGLF.Graphics.Models;
using SGLF.Graphics.Shaders;

namespace SGLF.Graphics
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    public abstract class Renderer
    {
        // ==================================================================================================
        // Constants
        // ==================================================================================================

        public int MaxLightsPerBatch = 8;

        // ==================================================================================================
        // Sub-Structures
        // ==================================================================================================

        /// <summary>
        ///
        /// </summary>
        protected struct SceneInfo
        {
            public UniformSet UniformSet;
            public Uniform<float> Time;            // Elapsed time.
            public Uniform<Matrix4> Model;         // Current model matrix.
            public Uniform<Matrix4> View;          // Current view matrix.
            public Uniform<Matrix4> Proj;          // Current projection matrix.
            public Uniform<Matrix4> ViewProj;      // View + Projection.
            public Uniform<Matrix4> ModelView;     // Model + View.
            public Uniform<Matrix4> ModelViewProj; // Model + View + Projection.
            public Uniform<Matrix3> Normal;        // Model + View normal matrix.
            public Uniform<Matrix3> ViewNormal;    // View normal matrix.
            public Uniform<float> NearClipPlane;   // Camera's near clip plane.
            public Uniform<float> FarClipPlane;    // Camera's far clip plane.
            public Uniform<Vector2> PixelSize;     // Size of a screen pixel in normalized space.
            public Uniform<Vector3> ClearColor;    // Screen clear color.
            public Quaternion ViewNormalQ;         // View normal quaternion.
        };

        /// <summary>
        ///
        /// </summary>
        protected struct AmbientLightInfo
        {
            public UniformSet UniformSet;
            public Uniform<Vector3> Sky;           // Sky color. Linear space.
            public Uniform<Vector3> Ground;        // Ground color. Linear space.
            public Uniform<Vector3> Up;            // World up in view space.
        };

        /// <summary>
        ///
        /// </summary>
        protected struct LightInfo
        {
            public Uniform<Vector4> Position;      // The position of the light. View space.
            public Uniform<Vector3> Normal;        // The normal direction of the light (Directional, Spot, Area). View space.
            public Uniform<Vector3> Tangent;       // The tangent direction of the light (Area). View Space.
            public Uniform<Vector3> Irradiance;    // The irradiance term. Linear space.
            public Uniform<Vector4> Parameters;    // Specific parameters, dependant of light type.
        };

        /// <summary>
        ///
        /// </summary>
        public struct RenderLight
        {
            public Light Light;
            public Vector4 Position;
            public Vector3 Normal;
            public Vector3 Tangent;
            public Vector3 Irradiance;
            public Vector4 Parameters;
            public BoundingSphere BoundingSphere;
        }

        /// <summary>
        ///
        /// </summary>
        public struct RenderEntity
        {
            public Model Model;
            public Matrix4 ModelMatrix;
            public Matrix4 ModelViewMatrix;
            public Matrix4 ModelViewProjMatrix;
            public Matrix3 NormalMatrix;
            public BoundingSphere BoundingSphere;
        }

        // ==================================================================================================
        // Members
        // ==================================================================================================

        protected Frustrum Frustrum;
        protected Rectangle Viewport;

        protected List<RenderEntity> RenderEntities;
        protected List<RenderLight> RenderLights;

        protected SceneInfo UScene;
        protected AmbientLightInfo UAmbientLight;
        protected UniformSet ULightUniformSet;
        protected Uniform<int> UNumLights;
        protected LightInfo[] ULight;

        // ==================================================================================================
        // Properties
        // ==================================================================================================

        /// <summary>
        /// Gets or sets the width of the screen.
        /// </summary>
        /// <value>
        /// The width of the screen.
        /// </value>
        public int Width
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the height of the screen.
        /// </summary>
        /// <value>
        /// The height of the screen.
        /// </value>
        public int Height
        {
            get;
            set;
        }

        /// <summary>
        /// Gets the cameras.
        /// </summary>
        /// <value>
        /// The cameras.
        /// </value>
        public HashSet<Camera> Cameras
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets or sets the color of the sky.
        /// </summary>
        /// <value>
        /// The color of the sky.
        /// </value>
        public Vector3 SkyColor
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the color of the ground.
        /// </summary>
        /// <value>
        /// The color of the ground.
        /// </value>
        public Vector3 GroundColor
        {
            get;
            set;
        }

        /// <summary>
        /// Gets the lights.
        /// </summary>
        /// <value>
        /// The lights.
        /// </value>
        public HashSet<Light> Lights
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the entities.
        /// </summary>
        /// <value>
        /// The entities.
        /// </value>
        public HashSet<Entity> Entities
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="Renderer"/> class.
        /// </summary>
        public Renderer()
        {
            // Initialize scene.
            Cameras = new HashSet<Camera>();
            Lights = new HashSet<Light>();
            Entities = new HashSet<Entity>();

            // Initialize render lists.
            RenderLights = new List<RenderLight>();
            RenderEntities = new List<RenderEntity>();

            // Initialize render uniforms.
            UScene = new SceneInfo();
            UScene.UniformSet = new UniformSet
            (
                UScene.Time = new Uniform<float>("uScene.Time"),
                UScene.Model = new Uniform<Matrix4>("uScene.Model"),
                UScene.View = new Uniform<Matrix4>("uScene.View"),
                UScene.Proj = new Uniform<Matrix4>("uScene.Proj"),
                UScene.ViewProj = new Uniform<Matrix4>("uScene.ViewProj"),
                UScene.ModelView = new Uniform<Matrix4>("uScene.ModelView"),
                UScene.ModelViewProj = new Uniform<Matrix4>("uScene.ModelViewProj"),
                UScene.Normal = new Uniform<Matrix3>("uScene.Normal"),
                UScene.ViewNormal = new Uniform<Matrix3>("uScene.ViewNormal"),
                UScene.NearClipPlane = new Uniform<float>("uScene.NearClipPlane"),
                UScene.FarClipPlane = new Uniform<float>("uScene.FarClipPlane"),
                UScene.PixelSize = new Uniform<Vector2>("uScene.PixelSize"),
                UScene.ClearColor = new Uniform<Vector3>("uScene.ClearColor")
            );

            UAmbientLight = new AmbientLightInfo();
            UAmbientLight.UniformSet = new UniformSet
            (
                UAmbientLight.Sky = new Uniform<Vector3>("uAmbientLight.Ground"),
                UAmbientLight.Ground = new Uniform<Vector3>("uAmbientLight.Sky"),
                UAmbientLight.Up = new Uniform<Vector3>("uAmbientLight.Up")
            );

            ULight = new LightInfo[MaxLightsPerBatch];
            ULightUniformSet = new UniformSet
            (
                UNumLights = new Uniform<int>("uNumLights")
            );

            for (int i = 0; i < MaxLightsPerBatch; i++)
            {
                ULightUniformSet.Add(ULight[i].Position = new Uniform<Vector4>("uLight[" + i + "].Position"));
                ULightUniformSet.Add(ULight[i].Normal = new Uniform<Vector3>("uLight[" + i + "].Normal"));
                ULightUniformSet.Add(ULight[i].Tangent = new Uniform<Vector3>("uLight[" + i + "].Tangent"));
                ULightUniformSet.Add(ULight[i].Irradiance = new Uniform<Vector3>("uLight[" + i + "].Irradiance"));
                ULightUniformSet.Add(ULight[i].Parameters = new Uniform<Vector4>("uLight[" + i + "].Parameters"));
            }
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        /// <summary>
        /// Adds the specified camera.
        /// </summary>
        /// <param name="camera">The camera.</param>
        public void Add(Camera camera)
        {
            if (Cameras.Add(camera))
                OnCameraAdded(camera);
        }

        /// <summary>
        /// Adds the specified light.
        /// </summary>
        /// <param name="light">The light.</param>
        public void Add(Light light)
        {
            if (Lights.Add(light))
                OnLightAdded(light);
        }

        /// <summary>
        /// Adds the specified entity.
        /// </summary>
        /// <param name="entity">The entity.</param>
        public void Add(Entity entity)
        {
            if (Entities.Add(entity))
                OnEntityAdded(entity);
        }

        /// <summary>
        /// Removes the specified camera.
        /// </summary>
        /// <param name="camera">The camera.</param>
        public void Remove(Camera camera)
        {
            if (Cameras.Remove(camera))
                OnCameraRemoved(camera);
        }

        /// <summary>
        /// Removes the specified light.
        /// </summary>
        /// <param name="light">The light.</param>
        public void Remove(Light light)
        {
            if (Lights.Remove(light))
                OnLightRemoved(light);
        }

        /// <summary>
        /// Removes the specified entity.
        /// </summary>
        /// <param name="entity">The entity.</param>
        public void Remove(Entity entity)
        {
            if (Entities.Remove(entity))
                OnEntityRemoved(entity);
        }

        /// <summary>
        /// Called when [camera added].
        /// </summary>
        /// <param name="camera">The camera.</param>
        public virtual void OnCameraAdded(Camera camera)
        {
        }

        /// <summary>
        /// Called when [camera removed].
        /// </summary>
        /// <param name="camera">The camera.</param>
        public virtual void OnCameraRemoved(Camera camera)
        {
        }

        /// <summary>
        /// Called when [light added].
        /// </summary>
        /// <param name="light">The light.</param>
        public virtual void OnLightAdded(Light light)
        {
        }

        /// <summary>
        /// Called when [light removed].
        /// </summary>
        /// <param name="light">The light.</param>
        public virtual void OnLightRemoved(Light light)
        {
        }

        /// <summary>
        /// Called when [entity added].
        /// </summary>
        /// <param name="entity">The entity.</param>
        public virtual void OnEntityAdded(Entity entity)
        {
        }

        /// <summary>
        /// Called when [entity removed].
        /// </summary>
        /// <param name="entity">The entity.</param>
        public virtual void OnEntityRemoved(Entity entity)
        {
        }

        /// <summary>
        /// Renders this instance.
        /// </summary>
        public abstract void Render();

        /// <summary>
        /// Cameras the bind.
        /// </summary>
        /// <param name="camera">The camera.</param>
        public void CameraBind(Camera camera)
        {
            // Get camera viewport.
            Viewport = camera.GetViewportRectangle(Width, Height);
            float aspectRatio = (float) Viewport.Width / Viewport.Height;

            // Setup scene info.
            UScene.Proj.Value = camera.GetProjectionMatrix(aspectRatio);
            UScene.View.Value = camera.Transform.InvMatrix;
            UScene.ViewProj.Value = UScene.View.Value * UScene.Proj.Value;
            UScene.ViewNormalQ = Quaternion.Invert(camera.Transform.Rotation);
            UScene.ViewNormal.Value = Matrix3.CreateFromQuaternion(UScene.ViewNormalQ);
            UScene.NearClipPlane.Value = camera.NearClipPlane;
            UScene.FarClipPlane.Value = camera.FarClipPlane;
            UScene.PixelSize.Value = new Vector2(1.0f / Viewport.Width, 1.0f / Viewport.Height);
            UScene.ClearColor.Value = Vector3Helper.FromColor(camera.ClearColor);

            // Setup frustrum and viewport.
            Frustrum = new Frustrum(UScene.ViewProj.Value);

            // Cull lights and entities.
            EntityCull();
            LightCull();
        }

        /// <summary>
        /// Cameras the unbind.
        /// </summary>
        public void CameraUnbind()
        {
            // Setup scene info.
            UScene.Proj.Value = Matrix4.Identity;
            UScene.View.Value = Matrix4.Identity;
            UScene.ViewProj.Value = Matrix4.Identity;
            UScene.ViewNormalQ = Quaternion.Identity;
            UScene.ViewNormal.Value = Matrix3.Identity;
            UScene.PixelSize.Value = Vector2.One;

            // Setup viewport.
            GL.Viewport(0, 0, Width, Height);
        }

        /// <summary>
        /// Ambients the bind.
        /// </summary>
        public void AmbientBind()
        {
            UAmbientLight.Ground.Value = GroundColor;
            UAmbientLight.Sky.Value = SkyColor;
            UAmbientLight.Up.Value = Vector3.Transform(Vector3.UnitY, UScene.ViewNormalQ);
        }

        /// <summary>
        /// Ambients the unbind.
        /// </summary>
        public void AmbientUnbind()
        {
            UAmbientLight.Ground.Value = Vector3.Zero;
            UAmbientLight.Sky.Value = Vector3.Zero;
            UAmbientLight.Up.Value = Vector3.Zero;
        }

        /// <summary>
        /// Lights the cull.
        /// </summary>
        public void LightCull()
        {
            RenderLights.Clear();

            foreach (Light light in Lights)
            {
                if ((light.LightType == 0) || Frustrum.TestSphere(light.BoundingSphere))
                    RenderLights.Add(LightSetup(light));
            }
        }

        /// <summary>
        /// Lights the setup.
        /// </summary>
        /// <param name="light">The light.</param>
        /// <returns></returns>
        public RenderLight LightSetup(Light light)
        {
            // Light position to view space.
            Vector4 position = new Vector4(light.Transform.Position, 1.0f);
            position = Vector4.Transform(position, UScene.View.Value);
            position.W = light.LightType;

            // Light normal to view space.
            Vector3 normal = light.Transform.Forward;
            normal = Vector3.Transform(normal, UScene.ViewNormalQ);

            // Light tangent to view space.
            Vector3 tangent = light.Transform.Right;
            tangent = Vector3.Transform(tangent, UScene.ViewNormalQ);

            // Done, build visibile light instance.
            RenderLight visibleLight = new RenderLight();
            visibleLight.Position = position;
            visibleLight.Normal = normal;
            visibleLight.Tangent = tangent;
            visibleLight.Irradiance = light.Irradiance;
            visibleLight.Parameters = new Vector4(light.Radius, light.Parameters.X, light.Parameters.Y, light.Parameters.Z);
            visibleLight.BoundingSphere = light.BoundingSphere;
            return visibleLight;
        }

        /// <summary>
        /// Lights the bind.
        /// </summary>
        /// <param name="l">The calculate.</param>
        /// <param name="s">The arguments.</param>
        public void LightBind(int l, int s)
        {
            ULight[s].Position.Value = RenderLights[l].Position;
            ULight[s].Normal.Value = RenderLights[l].Normal;
            ULight[s].Tangent.Value = RenderLights[l].Tangent;
            ULight[s].Irradiance.Value = RenderLights[l].Irradiance;
            ULight[s].Parameters.Value = RenderLights[l].Parameters;
        }

        /// <summary>
        /// Lights the unbind.
        /// </summary>
        /// <param name="s">The arguments.</param>
        public void LightUnbind(int s)
        {
            ULight[s].Position.Value = Vector4.Zero;
            ULight[s].Normal.Value = Vector3.Zero;
            ULight[s].Tangent.Value = Vector3.Zero;
            ULight[s].Irradiance.Value = Vector3.Zero;
            ULight[s].Parameters.Value = Vector4.Zero;
        }

        /// <summary>
        /// Entities the cull.
        /// </summary>
        public void EntityCull()
        {
            RenderEntities.Clear();

            foreach (Entity entity in Entities)
            {
                if (entity.Model == null)
                    continue;

                if (Frustrum.TestSphere(entity.BoundingSphere))
                {
                    RenderEntity renderEntity = new RenderEntity();
                    renderEntity.Model = entity.Model;
                    renderEntity.ModelMatrix = entity.Transform.Matrix;
                    renderEntity.ModelViewMatrix = renderEntity.ModelMatrix * UScene.View.Value;
                    renderEntity.ModelViewProjMatrix = renderEntity.ModelViewMatrix * UScene.Proj.Value;
                    renderEntity.NormalMatrix = Matrix3.CreateFromQuaternion(UScene.ViewNormalQ * entity.Transform.Rotation);
                    renderEntity.BoundingSphere = entity.BoundingSphere;
                    RenderEntities.Add(renderEntity);
                }
            }
        }

        /// <summary>
        /// Entities the bind.
        /// </summary>
        /// <param name="e">The decimal.</param>
        public void EntityBind(int e)
        {
            UScene.Model.Value = RenderEntities[e].ModelMatrix;
            UScene.ModelView.Value = RenderEntities[e].ModelViewMatrix;
            UScene.ModelViewProj.Value = RenderEntities[e].ModelViewProjMatrix;
            UScene.Normal.Value = RenderEntities[e].NormalMatrix;
        }

        /// <summary>
        /// Entities the unbind.
        /// </summary>
        public void EntityUnbind()
        {
            UScene.Model.Value = Matrix4.Identity;
            UScene.ModelView.Value = UScene.View.Value;
            UScene.ModelViewProj.Value = UScene.ViewProj.Value;
            UScene.Normal.Value = UScene.ViewNormal.Value;
        }
    }
}