﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using SGLF.Graphics.Buffers;
using SGLF.Graphics.Cameras;
using SGLF.Graphics.Materials;
using SGLF.Graphics.Models;
using SGLF.Graphics.Shaders;
using SGLF.Graphics.Textures;
using SGLF.Graphics.Vertices;

namespace SGLF.Graphics
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    public sealed class RendererDeferred : Renderer
    {
        // ==================================================================================================
        // Members
        // ==================================================================================================

        private Dictionary<Camera, GBuffer> CameraGBuffers;

        private GBuffer CurrentGBuffer;
        private Colorbuffer CurrentColorBuffer;

        private Mesh MeshQuad;

        private UberProgram UberClearGBuffer;
        private Program ProgramClearGBuffer;

        private UberProgram UberCompositeGBuffer;
        private Program ProgramAmbientPass;
        private Program ProgramLightPass;

        private UberProgram UberBlit;
        private Program ProgramBlit;

        private UberProgram UberSSR;
        private Program ProgramSSR;

        private UberProgram UberSSAO;
        private Program ProgramSSAO;

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        public RendererDeferred()
        {
            CameraGBuffers = new Dictionary<Camera, GBuffer>();
            CreateScreenQuad();
            CreatePrograms();
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        public override void OnCameraAdded(Camera camera)
        {
            CameraGBuffers[camera] = new GBuffer();
        }

        public override void OnCameraRemoved(Camera camera)
        {
            CameraGBuffers[camera].Dispose();
            CameraGBuffers.Remove(camera);
        }

        public override void Render()
        {
            // Clear screen.
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

            // Enable face culling.
            GL.Enable(EnableCap.CullFace);
            GL.CullFace(CullFaceMode.Back);

            // Render each camera.
            foreach (Camera camera in Cameras)
            {
                // Bind camera.
                CameraBind(camera);
                CameraSetupBuffers(camera);

                // Geometry pass.
                SetLocalViewport();
                BindGBufferForWrite();
                ClearPass();
                GeometryPass();
                UnbindGBufferForWrite();

                // Acquire color buffer.
                Colorbuffer colorbuffer1 = GraphicsManager.Colorbuffer.Acquire(Viewport.Width, Viewport.Height, (InternalFormat) All.Rgba16f);
                Colorbuffer colorbuffer2 = GraphicsManager.Colorbuffer.Acquire(Viewport.Width, Viewport.Height, (InternalFormat) All.Rgba16f);

                // Composition pass.
                BindColorBufferForWrite(colorbuffer1);
                BindGBufferForRead();
                BindScreenQuad();

                AmbientLightPass();
                LightPass();

                UnbindColorBufferForWrite(colorbuffer1);

                BindColorBufferForRead(colorbuffer1);
                BindColorBufferForWrite(colorbuffer2);
                PostProcessSSAO();
                UnbindColorBufferForWrite(colorbuffer2);

                BindColorBufferForRead(colorbuffer2);
                BindColorBufferForWrite(colorbuffer1);
                PostProcessSSR();
                UnbindColorBufferForWrite(colorbuffer1);

                UnbindGBufferForRead();

                // Render to screen pass.
                SetScreenViewport();
                BindColorBufferForRead(colorbuffer1);
                CopyPass();

                // Done.
                UnbindScreenQuad();
                UnbindColorBufferForRead(colorbuffer1);
                CameraUnbind();
                GraphicsManager.Colorbuffer.Release(colorbuffer1);
            }

            GraphicsManager.Colorbuffer.Recollect();
        }

        private void SetLocalViewport()
        {
            GL.Viewport(0, 0, Viewport.Width, Viewport.Height);
        }

        private void SetScreenViewport()
        {
            GL.Viewport(Viewport.X, Viewport.Y, Viewport.Width, Viewport.Height);
        }

        private void GeometryPass()
        {
            for (int i = 0; i < RenderEntities.Count; i++)
            {
                EntityBind(i);

                // Per each mesh...
                for (int j = 0; j < RenderEntities[i].Model.Meshes.Count; j++)
                {
                    Mesh mesh = RenderEntities[i].Model.Meshes[j];
                    Material material = RenderEntities[i].Model.Materials[j];
                    IndexBuffer indices = mesh.IndexBuffer;
                    VertexArrayObject vertexArray = mesh.VertexArray;

                    // Bind mesh.
                    GraphicsState.BindVertexArray(vertexArray);
                    GraphicsState.BindBuffer(indices);

                    // Bind material.
                    Program program = material.Program.Get("DEFERRED");
                    GraphicsState.BindProgram(program);
                    program.SetUniform(UScene.UniformSet);
                    program.SetUniform("uMaterial.", material.UniformSet);

                    // Bind textures.
                    material.UniformSet.ForEachUniform<UniformSampler>(u =>
                        GraphicsState.BindTexture(u.TextureUnit, u.Texture));

                    // Render mesh.
                    GL.DrawElements(indices.Primitive, indices.IndexCount, indices.IndexType, 0);
                }
            }
        }

        private void LightPass()
        {
            GraphicsState.BindProgram(ProgramLightPass);
            ProgramLightPass.SetUniform(UScene.UniformSet);
            GL.BlendFunc(BlendingFactorSrc.One, BlendingFactorDest.One);
            GL.Enable(EnableCap.Blend);

            for (int i = 0; i < RenderLights.Count; i += MaxLightsPerBatch)
            {
                UNumLights.Value = Math.Min(RenderLights.Count - i, MaxLightsPerBatch);

                for (int j = 0; j < UNumLights.Value; j++)
                    LightBind(i + j, j);

                ProgramLightPass.SetUniform(ULightUniformSet);
                RenderScreenQuad();
            }

            GL.Disable(EnableCap.Blend);
        }

        private void AmbientLightPass()
        {
            AmbientBind();
            GraphicsState.BindProgram(ProgramAmbientPass);
            ProgramAmbientPass.SetUniform(UScene.UniformSet);
            ProgramAmbientPass.SetUniform(UAmbientLight.UniformSet);
            RenderScreenQuad();
        }

        private void ClearPass()
        {
            GL.DepthFunc(DepthFunction.Always);
            GraphicsState.BindProgram(ProgramClearGBuffer);
            ProgramClearGBuffer.SetUniform(UScene.UniformSet);
            BindScreenQuad();
            RenderScreenQuad();
            UnbindScreenQuad();
            GL.DepthFunc(DepthFunction.Lequal);
        }

        private void PostProcessSSAO()
        {
            GraphicsState.BindProgram(ProgramSSAO);
            ProgramSSAO.SetUniform(UScene.UniformSet);
            RenderScreenQuad();
        }

        private void PostProcessSSR()
        {
            GraphicsState.BindProgram(ProgramSSR);
            ProgramSSR.SetUniform(UScene.UniformSet);
            RenderScreenQuad();
        }

        private void CopyPass()
        {
            GraphicsState.BindProgram(ProgramBlit);
            ProgramBlit.SetUniform(UScene.UniformSet);
            RenderScreenQuad();
        }

        private void CameraSetupBuffers(Camera camera)
        {
            CurrentGBuffer = CameraGBuffers[camera];
            CurrentGBuffer.Resize(Viewport.Width, Viewport.Height);
        }

        private void CreatePrograms()
        {
            UberClearGBuffer = GraphicsManager.Programs.Load("Render.Deferred.Clear.prog");
            UberCompositeGBuffer = GraphicsManager.Programs.Load("Render.Deferred.Composite.prog");
            UberBlit = GraphicsManager.Programs.Load("Render.Blit.prog");
            UberSSR = GraphicsManager.Programs.Load("Effect.SSR.prog");
            UberSSAO = GraphicsManager.Programs.Load("Effect.SSAO.prog");

            ProgramClearGBuffer = UberClearGBuffer.Get();
            ProgramAmbientPass = UberCompositeGBuffer.Get("AMBIENT_PASS");
            ProgramLightPass = UberCompositeGBuffer.Get();
            ProgramBlit = UberBlit.Get();
            ProgramSSR = UberSSR.Get();
            ProgramSSAO = UberSSAO.Get();
        }

        private void CreateScreenQuad()
        {
            MeshQuad = new Mesh(VertexFormat.FromType<VertexPU>(), 4, PrimitiveType.Triangles, 6,
                                  DrawElementsType.UnsignedInt, false);

            MeshQuad.VertexBuffer.SetData<VertexPU>(0, new VertexPU[]
            {
                new VertexPU(new Vector3(-1, -1, 0), new Vector2(0, 0)),
                new VertexPU(new Vector3( 1, -1, 0), new Vector2(1, 0)),
                new VertexPU(new Vector3( 1,  1, 0), new Vector2(1, 1)),
                new VertexPU(new Vector3(-1,  1, 0), new Vector2(0, 1))
            });

            MeshQuad.IndexBuffer.SetData(0, new int[]
            {
                0, 1, 2,
                0, 2, 3
            });
        }

        private void BindScreenQuad()
        {
            GraphicsState.BindVertexArray(MeshQuad.VertexArray);
            GraphicsState.BindBuffer(MeshQuad.IndexBuffer);
        }

        private void RenderScreenQuad()
        {
            GL.DrawElements(MeshQuad.IndexBuffer.Primitive, MeshQuad.IndexBuffer.IndexCount,
                            MeshQuad.IndexBuffer.IndexType, 0);
        }

        private void UnbindScreenQuad()
        {
            GraphicsState.UnbindBuffer(MeshQuad.IndexBuffer.Target);
            GraphicsState.UnbindVertexArray();
        }

        public void BindGBufferForWrite()
        {
            GraphicsState.BindFramebuffer(CurrentGBuffer.Framebuffer);
        }

        public void UnbindGBufferForWrite()
        {
            GraphicsState.UnbindFramebuffer(CurrentGBuffer.Framebuffer.Target);
        }

        public void BindGBufferForRead()
        {
            GraphicsState.BindTexture(TextureUnit.Texture1, CurrentGBuffer.Color1);
            GraphicsState.BindTexture(TextureUnit.Texture2, CurrentGBuffer.Color2);
            GraphicsState.BindTexture(TextureUnit.Texture3, CurrentGBuffer.Color3);
            GraphicsState.BindTexture(TextureUnit.Texture4, CurrentGBuffer.Color4);
            GraphicsState.BindTexture(TextureUnit.Texture5, CurrentGBuffer.Depth);
        }

        public void UnbindGBufferForRead()
        {
            GraphicsState.UnbindTexture(TextureUnit.Texture1);
            GraphicsState.UnbindTexture(TextureUnit.Texture2);
            GraphicsState.UnbindTexture(TextureUnit.Texture3);
            GraphicsState.UnbindTexture(TextureUnit.Texture4);
            GraphicsState.UnbindTexture(TextureUnit.Texture5);
        }

        public void BindColorBufferForWrite(Colorbuffer colorbuffer)
        {
            GraphicsState.BindFramebuffer(colorbuffer.Framebuffer);
        }

        public void UnbindColorBufferForWrite(Colorbuffer colorbuffer)
        {
            GraphicsState.UnbindFramebuffer(colorbuffer.Framebuffer.Target);
        }

        public void BindColorBufferForRead(Colorbuffer colorbuffer)
        {
            GraphicsState.BindTexture(TextureUnit.Texture0, colorbuffer.Color);
        }

        public void UnbindColorBufferForRead(Colorbuffer colorbuffer)
        {
            GraphicsState.UnbindTexture(TextureUnit.Texture0);
        }
    }
}