﻿using System.Collections.Generic;
using OpenTK;
using OpenTK.Graphics.OpenGL;

namespace SGLF.Graphics.Shaders
{
    public class Program : GraphicsNamedResource
    {
        private Dictionary<string, int> locations;
        private Dictionary<ShaderType, Shader> shaders;

        public IEnumerable<Shader> Shaders
        {
            get
            {
                return shaders.Values;
            }
        }

        public bool IsLinked
        {
            get
            {
                int isLinked;
                GL.GetProgram(Handle, ProgramParameter.LinkStatus, out isLinked);
                return (isLinked == 1);
            }
        }

        public string InfoLog
        {
            get
            {
                return GL.GetProgramInfoLog((int) Handle);
            }
        }

        public Program()
        {
            Handle = GL.CreateProgram();
            shaders = new Dictionary<ShaderType, Shader>();
            locations = new Dictionary<string, int>();
        }

        protected override void Dispose(bool disposeManaged, bool disposeUnmanaged)
        {
            if (disposeUnmanaged)
                GL.DeleteProgram(Handle);

            base.Dispose(disposeManaged, disposeUnmanaged);
        }

        public void Attach(Shader shader)
        {
            GL.AttachShader(Handle, shader.Handle);
            shaders[shader.Type] = shader;
        }

        public void Detach(Shader shader)
        {
            if (FindAttachment(shader.Type) == shader)
            {
                GL.DetachShader(Handle, shader.Handle);
                shaders[shader.Type] = null;
            }
        }

        public Shader FindAttachment(ShaderType type)
        {
            Shader shader = null;
            shaders.TryGetValue(type, out shader);
            return shader;
        }

        public int FindUniform(string name)
        {
            int location;

            if (!locations.TryGetValue(name, out location))
            {
                location = GL.GetUniformLocation(Handle, name);
                locations[name] = location;
            }

            return location;
        }

        public void Link()
        {
            GL.LinkProgram(Handle);
            locations.Clear();

            if (!IsLinked)
                throw new GraphicsException(InfoLog);
        }

        public void SetUniform(int location, int value0)
        {
            if (location != -1)
                GL.Ext.ProgramUniform1(Handle, location, value0);
        }

        public void SetUniform(int location, int value0, int value1)
        {
            if (location != -1)
                GL.Ext.ProgramUniform2(Handle, location, value0, value1);
        }

        public void SetUniform(int location, int value0, int value1, int value2)
        {
            if (location != -1)
                GL.Ext.ProgramUniform3(Handle, location, value0, value1, value2);
        }

        public void SetUniform(int location, int value0, int value1, int value2, int value3)
        {
            if (location != -1)
                GL.Ext.ProgramUniform4(Handle, location, value0, value1, value2, value3);
        }

        public void SetUniform(int location, int[] values, int elements, int count)
        {
            if (location != -1)
            {
                switch (elements)
                {
                    case 1:
                        GL.Ext.ProgramUniform1(Handle, location, count, values);
                        break;

                    case 2:
                        GL.Ext.ProgramUniform2(Handle, location, count, values);
                        break;

                    case 3:
                        GL.Ext.ProgramUniform3(Handle, location, count, values);
                        break;

                    case 4:
                        GL.Ext.ProgramUniform4(Handle, location, count, values);
                        break;
                }
            }
        }

        public void SetUniform(int location, float value0)
        {
            if (location != -1)
                GL.Ext.ProgramUniform1(Handle, location, value0);
        }

        public void SetUniform(int location, float value0, float value1)
        {
            if (location != -1)
                GL.Ext.ProgramUniform2(Handle, location, value0, value1);
        }

        public void SetUniform(int location, float value0, float value1, float value2)
        {
            if (location != -1)
                GL.Ext.ProgramUniform3(Handle, location, value0, value1, value2);
        }

        public void SetUniform(int location, float value0, float value1, float value2, float value3)
        {
            if (location != -1)
                GL.Ext.ProgramUniform4(Handle, location, value0, value1, value2, value3);
        }

        public void SetUniform(int location, float[] values, int elements, int count)
        {
            if (location != -1)
            {
                switch (elements)
                {
                    case 1:
                        GL.Ext.ProgramUniform1(Handle, location, count, values);
                        break;

                    case 2:
                        GL.Ext.ProgramUniform2(Handle, location, count, values);
                        break;

                    case 3:
                        GL.Ext.ProgramUniform3(Handle, location, count, values);
                        break;

                    case 4:
                        GL.Ext.ProgramUniform4(Handle, location, count, values);
                        break;
                }
            }
        }

        public void SetUniform(int location, Vector2 value)
        {
            if (location != -1)
                GL.Ext.ProgramUniform2(Handle, location, value.X, value.Y);
        }

        public void SetUniform(int location, Vector3 value)
        {
            if (location != -1)
                GL.Ext.ProgramUniform3(Handle, location, value.X, value.Y, value.Z);
        }

        public void SetUniform(int location, Vector4 value)
        {
            if (location != -1)
                GL.Ext.ProgramUniform4(Handle, location, value.X, value.Y, value.Z, value.W);
        }

        public void SetUniform(int location, ref Matrix2 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix2(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix3 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix3(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix4 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix4(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix2x3 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix2x3(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix2x4 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix2x4(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix3x2 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix3x2(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix3x4 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix3x4(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix4x2 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix4x2(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, ref Matrix4x3 value)
        {
            if (location != -1)
            {
                unsafe
                {
                    fixed (float* ptr = &value.Row0.X)
                    {
                        GL.Ext.ProgramUniformMatrix4x3(Handle, location, 1, false, ptr);
                    }
                }
            }
        }

        public void SetUniform(int location, Matrix2 value)
        {
            Matrix2 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix3 value)
        {
            Matrix3 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix4 value)
        {
            Matrix4 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix2x3 value)
        {
            Matrix2x3 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix2x4 value)
        {
            Matrix2x4 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix3x2 value)
        {
            Matrix3x2 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix3x4 value)
        {
            Matrix3x4 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix4x2 value)
        {
            Matrix4x2 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(int location, Matrix4x3 value)
        {
            Matrix4x3 temp = value;
            SetUniform(location, ref temp);
        }

        public void SetUniform(string name, int value0)
        {
            SetUniform(FindUniform(name), value0);
        }

        public void SetUniform(string name, int value0, int value1)
        {
            SetUniform(FindUniform(name), value0, value1);
        }

        public void SetUniform(string name, int value0, int value1, int value2)
        {
            SetUniform(FindUniform(name), value0, value1, value2);
        }

        public void SetUniform(string name, int value0, int value1, int value2, int value3)
        {
            SetUniform(FindUniform(name), value0, value1, value2, value3);
        }

        public void SetUniform(string name, int[] values, int elements, int count)
        {
            SetUniform(FindUniform(name), values, elements, count);
        }

        public void SetUniform(string name, float value0)
        {
            SetUniform(FindUniform(name), value0);
        }

        public void SetUniform(string name, float value0, float value1)
        {
            SetUniform(FindUniform(name), value0, value1);
        }

        public void SetUniform(string name, float value0, float value1, float value2)
        {
            SetUniform(FindUniform(name), value0, value1, value2);
        }

        public void SetUniform(string name, float value0, float value1, float value2, float value3)
        {
            SetUniform(FindUniform(name), value0, value1, value2, value3);
        }

        public void SetUniform(string name, float[] values, int elements, int count)
        {
            SetUniform(FindUniform(name), values, elements, count);
        }

        public void SetUniform(string name, Vector2 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Vector3 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Vector4 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, ref Matrix2 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix3 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix4 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix2x3 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix2x4 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix3x2 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix3x4 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix4x2 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, ref Matrix4x3 value)
        {
            SetUniform(FindUniform(name), ref value);
        }

        public void SetUniform(string name, Matrix2 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix3 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix4 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix2x3 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix2x4 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix3x2 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix3x4 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix4x2 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(string name, Matrix4x3 value)
        {
            SetUniform(FindUniform(name), value);
        }

        public void SetUniform(UniformSet uniforms)
        {
            SetUniform("", uniforms);
        }

        public void SetUniform(string baseName, UniformSet uniforms)
        {
            foreach (var uniform in uniforms.Uniforms)
                uniform.SetUniform(this, baseName);
        }
    }
}