﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Extensions;

namespace SGLF.Graphics.Shaders
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    public sealed class UberProgram : GraphicsResource
    {
        // ==================================================================================================
        // Constants
        // ==================================================================================================

        private const int MaxAvailableOptions = 32;

        // ==================================================================================================
        // Sub-Structures
        // ==================================================================================================

        private struct UberVariant
        {
            public Program Program;
            public List<Shader> Shaders;

            public UberVariant(Program program, List<Shader> shaders)
            {
                Program = program;
                Shaders = shaders;
            }
        }

        // ==================================================================================================
        // Members
        // ==================================================================================================

        private List<string> Options;

        private Dictionary<int, UberVariant> Variants;

        private Dictionary<ShaderType, string> Sources;

        // ==================================================================================================
        // Properties
        // ==================================================================================================

        public bool Changed
        {
            get;
            private set;
        }

        public int Count
        {
            get
            {
                return Options.Count;
            }
        }

        public Program this[params string[] requestedOptions]
        {
            get
            {
                return Get(requestedOptions);
            }
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        public UberProgram()
        {
            Options = new List<string>();
            Variants = new Dictionary<int, UberVariant>();
            Sources = new Dictionary<ShaderType, string>();
        }

        protected override void Dispose(bool disposeManaged, bool disposeUnmanaged)
        {
            if (disposeUnmanaged)
                Clear();
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        public void SetOption(string name)
        {
            // Check arguments.
            if (string.IsNullOrEmpty(name))
                throw new ArgumentNullException("name");
            if (Options.Count == MaxAvailableOptions)
                throw new ArgumentOutOfRangeException();

            // Store option.
            if (!Options.Contains(name))
            {
                Options.Add(name);
                Changed = true;
            }
        }

        public void UnsetOption(string name)
        {
            // Check arguments.
            if (string.IsNullOrEmpty(name))
                throw new ArgumentNullException("name");

            // Remove option.
            Options.Remove(name);
            Changed = true;
        }

        public void SetShaderSource(ShaderType shaderType, string sourceCode)
        {
            // Check arguments.
            if (string.IsNullOrEmpty(sourceCode))
                throw new ArgumentNullException("sourceCode");

            // Store shader source.
            Sources[shaderType] = sourceCode;
            Changed = true;
        }

        public void UnsetShaderSource(ShaderType shaderType)
        {
            // Remove shader source.
            Sources.Remove(shaderType);
            Changed = true;
        }

        public void Generate()
        {
            for (int i = 0; i <= Options.Count * Options.Count; i++)
                Generate(i);

            Changed = false;
        }

        public void Generate(int index)
        {
            // Build macro block for current index.
            string macroBlock = IndexToMacroBlock(index);

            // Allocate program and shader objects.
            Program program = new Program();
            List<Shader> shaders = new List<Shader>();

            // Build shaders and link program.
            try
            {
                foreach (var typeSource in Sources)
                {
                    ShaderType type = typeSource.Key;
                    string sourceCode = macroBlock + TypeToMacroBlock(type) + typeSource.Value;

                    Shader shader = new Shader(type, sourceCode);

                    if (!shader.IsCompiled)
                        throw new GraphicsException("Unable to compile shader: " + shader.InfoLog);

                    program.Attach(shader);
                    shaders.Add(shader);
                }

                program.Link();

                if (!program.IsLinked)
                    throw new GraphicsException("Unable to link program: " + program.InfoLog);

                Variants.Add(index, new UberVariant(program, shaders));
            }
            catch (Exception ex)
            {
                program.Dispose();

                foreach (var shader in shaders)
                    shader.Dispose();

                throw ex;
            }
        }

        public void Clear()
        {
            foreach (var generated in Variants.Values)
            {
                generated.Program.Dispose();

                foreach (var shader in generated.Shaders)
                    shader.Dispose();
            }

            Variants.Clear();
            Changed = true;
        }

        public Program Get(params string[] requestedOptions)
        {
            if (Changed)
                throw new InvalidOperationException();

            int index = OptionsToIndex(requestedOptions);
            return Variants[index].Program;
        }

        private int OptionsToIndex(string[] requestedOptions)
        {
            int index = 0;

            for (int i = 0; i < Options.Count; i++)
                if (requestedOptions.Contains(Options[i]))
                    index = index.SetBit(i, true);

            return index;
        }

        private string IndexToMacroBlock(int index)
        {
            string macroBlock = "";

            for (int i = 0; i < Options.Count; i++)
                if (index.GetBit(i))
                    macroBlock += "#define OPTION_" + Options[i] + "\r\n";

            return macroBlock;
        }

        private string TypeToMacroBlock(ShaderType type)
        {
            switch (type)
            {
                case ShaderType.VertexShader:
                    return "#define SHADER_VERTEX\r\n";

                case ShaderType.FragmentShader:
                    return "#define SHADER_FRAGMENT\r\n";

                case ShaderType.GeometryShader:
                    return "#define SHADER_GEOMETRY\r\n";

                case ShaderType.TessControlShader:
                    return "#define SHADER_TESS_CONTROL\r\n";

                case ShaderType.TessEvaluationShader:
                    return "#define SHADER_TESS_EVAL\r\n";

                case ShaderType.ComputeShader:
                    return "#define SHADER_COMPUTE\r\n";

                default:
                    throw new ArgumentException("type");
            }
        }
    }
}