﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using OpenTK.Graphics.OpenGL;
using SGLF.Core.Helpers;

namespace SGLF.Graphics.Textures
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Represents an OpenGL framebuffer object.
    /// </summary>
    public class Framebuffer : GraphicsNamedResource
    {
        // ==================================================================================================
        // Properties
        // ==================================================================================================

        public int Width
        {
            get;
            private set;
        }

        public int Height
        {
            get;
            private set;
        }

        public FramebufferTarget Target
        {
            get;
            private set;
        }

        public FramebufferErrorCode Status
        {
            get
            {
                return (FramebufferErrorCode) GL.Ext.CheckNamedFramebufferStatus(Handle, Target);
            }
        }

        public Dictionary<FramebufferAttachment, object> Attachments
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="Framebuffer"/> class.
        /// </summary>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        public Framebuffer(int width, int height, FramebufferTarget target = FramebufferTarget.Framebuffer)
        {
            // Check size
            if (width <= 0)
                throw new ArgumentOutOfRangeException("width");
            if (height <= 0)
                throw new ArgumentOutOfRangeException("height");

            Width = width;
            Height = height;
            Target = target;
            Attachments = EnumHelper.CreateDictionary<FramebufferAttachment, object>();
            Handle = GL.GenFramebuffer();
        }

        /// <summary>
        /// Disposes the instance's resources.
        /// </summary>
        /// <param name="disposeManaged"><c>true</c> to release managed resources.</param>
        /// <param name="disposeUnmanaged"><c>true</c> to release unmanaged resources.</param>
        protected override void Dispose(bool disposeManaged, bool disposeUnmanaged)
        {
            if (disposeUnmanaged)
                GL.DeleteFramebuffer(Handle);

            base.Dispose(disposeManaged, disposeUnmanaged);
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        public void Attach(FramebufferAttachment attachment, Texture2D texture)
        {
            // Check parameters.
            if (texture == null)
                throw new ArgumentNullException("texture");
            if (texture.Disposed)
                throw new ObjectDisposedException("texture");
            if (texture.Width != Width || texture.Height != Height)
                throw new ArgumentException("texture");
            if (Attachments[attachment] == texture)
                return;

            // Attach texture.
            Attachments[attachment] = texture;
            GL.Ext.NamedFramebufferTexture2D(Handle, attachment, texture.Target, texture.Handle, 0);
        }

        public void Dettach(FramebufferAttachment attachment)
        {
            if (Attachments[attachment] == null)
                return;

            object attachedObject = Attachments[attachment];

            // Dettach.
            if (attachedObject is Texture2D)
            {
                Texture2D texture = (Texture2D) attachedObject;
                Attachments[attachment] = null;
                GL.Ext.NamedFramebufferTexture2D(Handle, attachment, texture.Target, 0, 0);
            }
        }

        public void Resize(int width, int height)
        {
            // Check size
            if (width <= 0)
                throw new ArgumentOutOfRangeException("width");
            if (height <= 0)
                throw new ArgumentOutOfRangeException("height");

            // Update size.
            if (Width != width || Height != height)
            {
                Width = width;
                Height = height;
                Discard();
            }
        }

        public void Discard()
        {
            // Discard framebuffer object.
            GL.DeleteFramebuffer(Handle);
            Handle = GL.GenFramebuffer();

            // Discard attachments.
            Attachments.Clear();
        }

        public void Realize()
        {
            // Get list of color attachments.
            DrawBufferMode[] colors = Attachments.Where
            (
                kv => (kv.Key >= FramebufferAttachment.ColorAttachment0) &&
                      (kv.Key <= FramebufferAttachment.ColorAttachment15) &&
                      (kv.Value != null)
            ).Select(
                a => (DrawBufferMode) a.Key
            ).ToArray();

            // Setup framebuffer.
            GL.Ext.FramebufferDrawBuffers(Handle, colors.Length, colors);
        }
    }
}