﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using OpenTK.Graphics.OpenGL;

namespace SGLF.Graphics.Textures
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    public class GBuffer : GraphicsResource
    {
        // +============================================================================+
        // | G-Buffer format                                                            |
        // +================+==========+===========+===========+===========+============+
        // | Name           | Format   | Red       | Green     | Blue      | Alpha      |
        // |················+··········+···········+···········+···········+············|
        // | 1 - Position   | RGBA32f  | PositionX | PositionY | PositionZ | AO         |
        // |················+··········+··········-+···········+···········+············|
        // | 2 - Normals    | RGBA16f  | NormalX   | NormalY   | NormalZ   |            |
        // |················+··········+···········+···········+···········+············|
        // | 3 - Albedo     | RGBA16f  | DiffuseR  | DiffuseG  | DiffuseB  |            |
        // |················+··········+···········+···········+···········+············|
        // | 4 - Specular   | RGBA16f  | Metallic  | Roughness | Gloss     | Emission   |
        // |················+··········+··········-+···········+···········+············|
        // | 5 - Depth      | Depth    |                     Depth                      |
        // +================+==========+================================================+

        // ==================================================================================================
        // Properties
        // ==================================================================================================

        public int Width
        {
            get;
            private set;
        }

        public int Height
        {
            get;
            private set;
        }

        public Framebuffer Framebuffer
        {
            get;
            private set;
        }

        public Texture2D Color1
        {
            get;
            private set;
        }

        public Texture2D Color2
        {
            get;
            private set;
        }

        public Texture2D Color3
        {
            get;
            private set;
        }

        public Texture2D Color4
        {
            get;
            private set;
        }

        public Texture2D Depth
        {
            get;
            private set;
        }

        public bool Created
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        public GBuffer()
        {
            Width = 0;
            Height = 0;
            Created = false;
        }

        public GBuffer(int width, int height)
        {
            Width = width;
            Height = height;
            Created = Create();
        }

        /// <summary>
        /// Disposes the instance's resources.
        /// </summary>
        /// <param name="disposeManaged"><c>true</c> to release managed resources.</param>
        /// <param name="disposeUnmanaged"><c>true</c> to release unmanaged resources.</param>
        protected override void Dispose(bool disposeManaged, bool disposeUnmanaged)
        {
            if (disposeUnmanaged)
                Discard();

            base.Dispose(disposeManaged, disposeUnmanaged);
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        private bool Create()
        {
            // Generate textures.
            Color1 = new Texture2D(Width, Height, (InternalFormat) All.Rgba32f, 0);
            Color1.MagFilter = TextureMagFilter.Nearest;
            Color1.MinFilter = TextureMinFilter.Nearest;
            Color1.WrapModeS = TextureWrapMode.Clamp;
            Color1.WrapModeT = TextureWrapMode.Clamp;

            Color2 = new Texture2D(Width, Height, (InternalFormat) All.Rgba16f, 0);
            Color2.MagFilter = TextureMagFilter.Nearest;
            Color2.MinFilter = TextureMinFilter.Nearest;
            Color2.WrapModeS = TextureWrapMode.Clamp;
            Color2.WrapModeT = TextureWrapMode.Clamp;

            Color3 = new Texture2D(Width, Height, (InternalFormat) All.Rgba16f, 0);
            Color3.MagFilter = TextureMagFilter.Nearest;
            Color3.MinFilter = TextureMinFilter.Nearest;
            Color3.WrapModeS = TextureWrapMode.Clamp;
            Color3.WrapModeT = TextureWrapMode.Clamp;

            Color4 = new Texture2D(Width, Height, (InternalFormat) All.Rgba16f, 0);
            Color4.MagFilter = TextureMagFilter.Nearest;
            Color4.MinFilter = TextureMinFilter.Nearest;
            Color4.WrapModeS = TextureWrapMode.Clamp;
            Color4.WrapModeT = TextureWrapMode.Clamp;

            Depth = new Texture2D(Width, Height, (InternalFormat) All.DepthComponent32, 0,
                                           PixelFormat.DepthComponent, PixelType.Int);
            Depth.MagFilter = TextureMagFilter.Nearest;
            Depth.MinFilter = TextureMinFilter.Nearest;
            Depth.WrapModeS = TextureWrapMode.Clamp;
            Depth.WrapModeT = TextureWrapMode.Clamp;

            // Generate framebuffer.
            Framebuffer = new Framebuffer(Width, Height);
            Framebuffer.Attach(FramebufferAttachment.ColorAttachment0, Color1);
            Framebuffer.Attach(FramebufferAttachment.ColorAttachment1, Color2);
            Framebuffer.Attach(FramebufferAttachment.ColorAttachment2, Color3);
            Framebuffer.Attach(FramebufferAttachment.ColorAttachment3, Color4);
            Framebuffer.Attach(FramebufferAttachment.DepthAttachment, Depth);
            Framebuffer.Realize();

            return true;
        }

        private void Discard()
        {
            if (Framebuffer != null)
                Framebuffer.Dispose();
            if (Color1 != null)
                Color1.Dispose();
            if (Color2 != null)
                Color2.Dispose();
            if (Color3 != null)
                Color3.Dispose();
            if (Color4 != null)
                Color4.Dispose();
            if (Depth != null)
                Depth.Dispose();
            Created = false;
        }

        public void Resize(int width, int height)
        {
            if (Width != width || Height != height)
            {
                Discard();

                Width = width;
                Height = height;

                Created = Create();
            }
        }
    }
}