﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using OpenTK.Graphics.OpenGL;

namespace SGLF.Graphics.Textures
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Represents an OpenGL renderbuffer object.
    /// </summary>
    public class Renderbuffer : GraphicsNamedResource
    {
        // ==================================================================================================
        // Properties
        // ==================================================================================================

        public int Width
        {
            get;
            private set;
        }

        public int Height
        {
            get;
            private set;
        }

        public PixelInternalFormat Format
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Constructor / Disposer
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="Framebuffer" /> class.
        /// </summary>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        /// <param name="format">The internal pixel format.</param>
        public Renderbuffer(int width, int height, PixelInternalFormat format)
        {
            // Check size
            if (width <= 0)
                throw new ArgumentOutOfRangeException("width");
            if (height <= 0)
                throw new ArgumentOutOfRangeException("height");

            Width = width;
            Height = height;
            Format = format;
            Handle = GL.GenRenderbuffer();
            GL.Ext.NamedRenderbufferStorage(Handle, Format, Width, Height);
        }

        /// <summary>
        /// Disposes the instance's resources.
        /// </summary>
        /// <param name="disposeManaged"><c>true</c> to release managed resources.</param>
        /// <param name="disposeUnmanaged"><c>true</c> to release unmanaged resources.</param>
        protected override void Dispose(bool disposeManaged, bool disposeUnmanaged)
        {
            if (disposeUnmanaged)
                GL.DeleteRenderbuffer(Handle);

            base.Dispose(disposeManaged, disposeUnmanaged);
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        /// <summary>
        /// Resizes the specified width.
        /// </summary>
        /// <param name="width">The width.</param>
        /// <param name="height">The height.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">
        /// width
        /// or
        /// height
        /// </exception>
        public void Resize(int width, int height, PixelInternalFormat format = 0)
        {
            // Check size
            if (width <= 0)
                throw new ArgumentOutOfRangeException("width");
            if (height <= 0)
                throw new ArgumentOutOfRangeException("height");

            // Fix default parameters.
            if (format == 0)
                format = Format;

            // Update size.
            if (Width != width || Height != height || Format != format)
            {
                Width = width;
                Height = height;
                Format = format;
                GL.Ext.NamedRenderbufferStorage(Handle, Format, Width, Height);
            }
        }
    }
}