﻿using System;
using System.Drawing;
using OpenTK.Graphics.OpenGL;
using BitmapData = System.Drawing.Imaging.BitmapData;
using BitmapLockMode = System.Drawing.Imaging.ImageLockMode;
using BitmapPixelFormat = System.Drawing.Imaging.PixelFormat;

namespace SGLF.Graphics.Textures
{
    public class Texture2D : TextureObject
    {
        public int Height
        {
            get;
            private set;
        }

        public int Width
        {
            get;
            private set;
        }

        public Texture2D(int width, int height, InternalFormat format, int numMipMaps,
                               PixelFormat allocPixelFormat = PixelFormat.Rgba,
                               PixelType alloxPixelType = PixelType.Byte)
            : base(format, TextureTarget.Texture2D, numMipMaps)
        {
            Width = width;
            Height = height;
            WrapModeS = TextureWrapMode.Repeat;
            WrapModeT = TextureWrapMode.Repeat;
            MagFilter = TextureMagFilter.Linear;
            MinFilter = TextureMinFilter.Linear;

            // TODO: Unhack allocation type and format, select from proper dictionary (unless a workaround is
            // found).
            GL.Ext.TextureImage2D(Handle, Target, NumMipMaps, (int) Format, Width, Height, 0,
                                  allocPixelFormat, alloxPixelType, IntPtr.Zero);
        }

        public void GenerateMipmaps()
        {
            GL.Ext.GenerateTextureMipmap(Handle, Target);
        }

        public void GetImage(PixelFormat pixelFormat, PixelType pixelType, IntPtr pixels, int level = 0)
        {
            // Check pixels.
            if (pixels == IntPtr.Zero)
                throw new ArgumentNullException("pixels");

            // Get image.
            GL.Ext.GetTextureImage(Handle, Target, level, pixelFormat, pixelType, pixels);
        }

        public Bitmap GetImage(int level = 0)
        {
            // Check level.
            if (level > NumMipMaps)
                throw new ArgumentException("level");

            // Determine width and height of bitmap.
            int width = Width >> level;
            int height = Height >> level;

            // Create bitmap and retrieve texture data.
            Bitmap bitmap = new Bitmap(width, height, BitmapPixelFormat.Format32bppArgb);
            BitmapData bitmapData = bitmap.LockBits(new Rectangle(0, 0, width, height), BitmapLockMode.WriteOnly, BitmapPixelFormat.Format32bppArgb);
            GetImage(PixelFormat.Bgra, PixelType.UnsignedByte, bitmapData.Scan0, level);
            bitmap.UnlockBits(bitmapData);
            bitmap.RotateFlip(RotateFlipType.RotateNoneFlipY);

            // Done.
            return bitmap;
        }

        public void SetImage(PixelFormat pixelFormat, PixelType pixelType, IntPtr pixels, int level = 0)
        {
            // Check pixels.
            if (pixels == IntPtr.Zero)
                throw new ArgumentNullException("pixels");

            // Set image.
            GL.Ext.TextureImage2D(Handle, Target, level, (int) Format, Width, Height, 0, pixelFormat,
                                  pixelType, pixels);
        }

        public void SetImage(Bitmap bitmap, int level = 0)
        {
            // Determine width and height of bitmap.
            int width = Width >> level;
            int height = Height >> level;

            // Check level / bitmap.
            if (level > NumMipMaps)
                throw new ArgumentException("level");
            if (bitmap == null)
                throw new ArgumentNullException("bitmap");
            if (bitmap.Width != width && bitmap.Height != height)
                throw new ArgumentOutOfRangeException("bitmap");

            // Lock bitmap and write texture data.
            BitmapData bitmapData = bitmap.LockBits(new Rectangle(0, 0, width, height), BitmapLockMode.ReadOnly, BitmapPixelFormat.Format32bppArgb);
            SetImage(PixelFormat.Bgra, PixelType.UnsignedByte, bitmapData.Scan0, level);
            bitmap.UnlockBits(bitmapData);
        }
    }
}