﻿// TODO: Header

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Dependencies
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;

namespace SGLF.Graphics.Vertices
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classes
    /////////////////////////////////////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Defines the layout and attributes of a vertex structure.
    /// </summary>
    public sealed class VertexFormat
    {
        // ==================================================================================================
        // Static Properties
        // ==================================================================================================

        /// <summary>
        /// The list of available vertex formats.
        /// </summary>
        public static List<VertexFormat> AvailableFormats
        {
            get;
            private set;
        }

        /// <summary>
        /// The dictionary containing the available vertex formats and their source vertex type.
        /// </summary>
        public static Dictionary<Type, VertexFormat> FormatCache
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Properties
        // ==================================================================================================

        /// <summary>
        /// The source vertex structure type.
        /// </summary>
        public Type VertexType
        {
            get;
            private set;
        }

        /// <summary>
        /// The attributes within the vertex structure.
        /// </summary>
        public VertexAttribute[] VertexAttributes
        {
            get;
            private set;
        }

        /// <summary>
        /// Size (in bytes) of the vertex structure.
        /// </summary>
        public int Stride
        {
            get;
            private set;
        }

        // ==================================================================================================
        // Static Constructor
        // ==================================================================================================

        /// <summary>
        /// Initializes the <see cref="VertexFormat"/> class.
        /// </summary>
        static VertexFormat()
        {
            AvailableFormats = new List<VertexFormat>();
            FormatCache = new Dictionary<Type, VertexFormat>();

            foreach (var type in Assembly.GetExecutingAssembly().GetTypes().Where
                                 (m => m.IsValueType && m.GetInterface("IVertex") != null))
            {
                IVertex instance = (IVertex) Activator.CreateInstance(type);
                AvailableFormats.Add(instance.Format);
                FormatCache.Add(type, instance.Format);
            }
        }

        // ==================================================================================================
        // Constructor
        // ==================================================================================================

        /// <summary>
        /// Initializes a new instance of the <see cref="VertexFormat"/> class.
        /// </summary>
        /// <param name="vertexType">Type of the vertex.</param>
        /// <param name="sizeInBytes">The size (in bytes) of the vertex.</param>
        /// <param name="attributes">The vertex attributes.</param>
        public VertexFormat(Type vertexType, params VertexAttribute[] attributes)
        {
            VertexType = vertexType;
            VertexAttributes = attributes.Clone() as VertexAttribute[];
            Stride = Marshal.SizeOf(vertexType);
        }

        // ==================================================================================================
        // Static Methods
        // ==================================================================================================

        /// <summary>
        /// Finds the smallest <see cref="VertexFormat"/> that contains the given semantics.
        /// </summary>
        /// <param name="semantics">The required semantics.</param>
        /// <returns>VertexFormat instance that satisfies the requirements. <c>null</c> if no
        /// format is available.</returns>
        public static VertexFormat FromSemantics(ICollection<VertexSemantics> semantics)
        {
            VertexFormat vertexFormat = null;

            foreach (var format in AvailableFormats)
            {
                bool match = true;

                foreach (var semantic in semantics)
                {
                    if (!format.ContainsSemantics(semantic))
                    {
                        match = false;
                        break;
                    }
                }

                if (match)
                {
                    if (vertexFormat == null)
                        vertexFormat = format;
                    else if (vertexFormat.Stride > format.Stride)
                        vertexFormat = format;
                }
            }

            return vertexFormat;
        }

        /// <summary>
        /// Retrieves the <see cref="VertexFormat"/> of the specified vertex type.
        /// </summary>
        /// <typeparam name="T"><see cref="IVertex"/> type.</typeparam>
        /// <returns>The VertexFormat of the requested type.</returns>
        public static VertexFormat FromType<T>()
            where T : IVertex
        {
            return FormatCache[typeof(T)];
        }

        /// <summary>
        /// Retrieves the <see cref="VertexFormat"/> of the specified vertex type.
        /// </summary>
        /// <param name="T">The attribute.</param>
        /// <returns></returns>
        public static VertexFormat FromType(Type T)
        {
            if (T == null)
                throw new ArgumentNullException("T");
            return FormatCache[T];
        }

        // ==================================================================================================
        // Methods
        // ==================================================================================================

        /// <summary>
        /// Determines whether the specified semantics is present within the vertex format.
        /// </summary>
        /// <param name="semantics">The semantics to find.</param>
        /// <returns><c>true</c> if the semantics is present, <c>false</c> otherwise.</returns>
        public bool ContainsSemantics(VertexSemantics semantics)
        {
            return VertexAttributes.Any(a => a.AttributeSemantics == semantics);
        }

        /// <summary>
        /// Finds the attribute that implements the specified semantics.
        /// </summary>
        /// <param name="semantics">The semantics.</param>
        /// <returns>A <see cref="VertexAttribute"/> instance, <c>null</c> if it doesn't exist.</returns>
        public VertexAttribute FindAttribute(VertexSemantics semantics)
        {
            return VertexAttributes.Where(a => a.AttributeSemantics == semantics).First();
        }
    }
}