#ifndef COMMON_INC
#define COMMON_INC
 
// ----------------------------------------------------------------------------
// Constants
// ----------------------------------------------------------------------------

#define PI                  3.14159265358   // PI
#define TWOPI               6.28318530718   // 2 PI
#define PIOVER2             1.57079632679   // PI / 2
#define PIOVER4             0.78539816339   // PI / 4
#define RECPI				0.31830988618	// 1 / PI

#define EPSILON             1e-10f
#define OMEGA               1e10f

#define MAX_LIGHTS          8               // Maximum 8 lights per pass.

#define LIGHT_DIRECTIONAL   0               // Directional light type.
#define LIGHT_POINT         1               // Point light type.
#define LIGHT_SPOT          2               // Spot light type.

// ----------------------------------------------------------------------------
// Structures
// ----------------------------------------------------------------------------

/*
** Contains the info of the current scene.
*/

struct SceneInfo
{
    float Time;                 // Elapsed time.

    mat4 Model;                 // Current model matrix.

    mat4 View;                  // Current view matrix.
    mat4 Proj;                  // Current projection matrix.
    mat4 ViewProj;              // View + Projection.

    mat4 ModelView;             // Model + View.
    mat4 ModelViewProj;         // Model + View + Projection.
    mat3 Normal;                // Model + View normal matrix.
    mat3 ViewNormal;            // View normal matrix.

    float NearClipPlane;        // Camera's near clip plane.
    float FarClipPlane;         // Camera's far clip plane.
	
	vec2 PixelSize;				// Size of a screen pixel in normalized space.
	
	vec3 ClearColor;            // Screen clear color.
};

/*
** Contains the info of a surface point.
*/
struct SurfaceInfo
{
    vec3 Position;              // The position of the surface. View space.
    vec3 Normal;                // The normal of the surface. View space.
	
    vec3 ViewDir;               // View direction. View space.
	
    vec3 BaseColor;             // The base color of the surface. Linear space.
    vec3 DiffuseColor;          // The surface's diffuse color. Linear space.
    vec3 SpecularColor;         // The surface's specular color. Linear space.
	
    float AmbientOcclusion;     // Amount of occlusion at the surface point. (0..1)
	
    float Metallic;             // Metalness term of the surface. (0..1)
    float Roughness;            // Roughness term of the surface. (0..1)
    float Specular;             // Specular intensity. (0..1)
    float Emission;             // Emissive intensity. (0..1)
};

/*
** Contains the info of the current ambient lighting.
*/
struct AmbientLightInfo
{
    vec3 Sky;                   // Sky color. Linear space.
    vec3 Ground;                // Ground color. Linear space.
    vec3 Up;                    // World up in view space.
};

/*
** Contains the info of a single light.
*/
struct LightInfo
{
    vec4 Position;              // The position of the light. View space.
    vec3 Normal;                // The normal direction of the light (Directional, Spot, Area). View space.
    vec3 Tangent;               // The tangent direction of the light (Area). View Space.
    vec3 Irradiance;            // The irradiance term. Linear space.
    vec4 Parameters;            // Specific parameters, dependant of light type:
                                // - Point (radius, sphere radius)
                                // - Spot (radius, min angle, max angle)
                                // - Plane (radius, plane width, plane height)
                                // - Tube (radius, tube length)
};

// ----------------------------------------------------------------------------
// Uniforms
// ----------------------------------------------------------------------------

uniform SceneInfo uScene;                       // Scene data.

uniform int uNumLights;
uniform LightInfo uLight[MAX_LIGHTS];           // Lights data.

uniform AmbientLightInfo uAmbientLight;         // Ambient light data.

// ----------------------------------------------------------------------------
// Functions
// ----------------------------------------------------------------------------

// --- Useful functions -------------------------------------------------------

/*
** Saturates the value in the range of 0..1
*/
float saturate(float value) { return clamp(value, 0, 1); }
vec2 saturate(vec2 value) { return clamp(value, 0, 1); }
vec3 saturate(vec3 value) { return clamp(value, 0, 1); }
vec4 saturate(vec4 value) { return clamp(value, 0, 1); }

/*
** Computes the square of the given value.
*/
float sqr(float value) { return value * value; }
float sqr(vec2 value) { return dot(value, value); }
float sqr(vec3 value) { return dot(value, value); }
float sqr(vec4 value) { return dot(value, value); }

// --- Color functions --------------------------------------------------------

/*
** Transforms from gamma space to linear space.
*/
vec4 GammaToLinear(vec4 color) { return color * color; }
vec3 GammaToLinear(vec3 color) { return color * color; }
vec2 GammaToLinear(vec2 color) { return color * color; }
float GammaToLinear(float color) { return color * color; }

/*
** Transforms from linear space to gamma space.
*/
vec4 LinearToGamma(vec4 color) { return sqrt(color); }
vec3 LinearToGamma(vec3 color) { return sqrt(color); }
vec2 LinearToGamma(vec2 color) { return sqrt(color); }
float LinearToGamma(float color) { return sqrt(color); }

// --- Geometric functions ----------------------------------------------------

/*
** Computes the distance to the given plane.
*/
float DistanceToPlane(vec3 point, vec4 plane)
{
    return dot(plane.xyz, point) + plane.w;
}

/*
** Computes the distance to the given sphere.
*/
float DistanceToSphere(vec3 point, vec4 sphere)
{
    return length(point - sphere.xyz) - sphere.w;
}

/*
** Computes the closest point on the plane.
*/
vec3 ClosestPointOnPlane(vec3 point, vec4 plane)
{
    return point - DistanceToPlane(point, plane) * plane.xyz;
}

/*
** Computes the closest point on the sphere.
*/
vec3 ClosestPointOnSphere(vec3 point, vec4 sphere)
{
    vec3 pointToSphere = point - sphere.xyz;
    return sphere.xyz + normalize(pointToSphere) * min(length(pointToSphere) - EPSILON, sphere.w);
}

/*
** Computes the closest point on the finite plane.
*/
vec3 ClosestPointOnFinitePlane(vec3 point, vec3 origin, vec4 plane, vec3 tangent, vec3 bitangent, vec2 size)
{
    vec3 closestPoint = ClosestPointOnPlane(point, plane);
    vec3 fromOrigin = closestPoint - origin;
    
    vec2 planeCoords = vec2(dot(fromOrigin, tangent), dot(fromOrigin, bitangent));
    planeCoords = clamp(planeCoords, -size, size);

    return origin + planeCoords.x * tangent + planeCoords.y * bitangent;    
}

/*
** Computes the intersection of a ray with a plane. Note: w = 1 if successful.
*/
vec4 RayPlaneIntersection(vec3 origin, vec3 direction, vec4 plane)
{
    float d = DistanceToPlane(origin, plane) / dot(plane.xyz, direction);
    return vec4(origin + direction * d, step(0, d));
}

/*
** Computes the intersection of a ray with a sphere. Note: w = 1 if successful.
*/
vec4 RaySphereIntersection(vec3 origin, vec3 direction, vec4 sphere)
{
    vec3 sphereToRay = origin - sphere.xyz;

    float a = dot(direction, direction);
    float b = 2 * dot(direction, sphereToRay);
    float c = dot(sphereToRay, sphereToRay) - (sphere.w * sphere.w);

    float discriminant = b * b - 4 * a * c;

    if (discriminant < EPSILON)
        return vec4(0);

    float sqrtDiscr = sqrt(discriminant);
    float t0 = (-b + sqrtDiscr) / (2.0 * a);
    float t1 = (-b - sqrtDiscr) / (2.0 * a);

    if (t0 < EPSILON) t0 = t1;
    if (t1 < EPSILON) t1 = t0;

    float distance = min(t0, t1);

    if (distance < EPSILON)
        return vec4(0);

    return vec4(origin + direction * distance, 1);
}

/*
** Computes the closest point of a ray to an sphere.
*/
vec3 RaySphereClosest(vec3 origin, vec3 direction, vec4 sphere)
{
	vec3 toCenter = sphere.xyz - origin;
	vec3 closestPointOnRay = dot(toCenter, direction) * direction;
	vec3 centerToRay = closestPointOnRay - toCenter;
	return origin + toCenter + centerToRay * saturate(sphere.w * inversesqrt(sqr(centerToRay)));
}

// --- Surface functions ------------------------------------------------------

/*
** Computes the surface's diffuse color from it's properties.
*/
vec3 SurfaceCalcDiffuse(vec3 baseColor, float metallic)
{
    return mix(baseColor, vec3(0), metallic);
}

/*
** Computes the surface's specular color from it's properties.
*/
vec3 SurfaceCalcSpecular(vec3 baseColor, float metallic, float specular)
{
    return mix(0.08 * specular.xxx, baseColor, metallic);
}

// --- GBuffer functions ------------------------------------------------------

/*
** Stores the surface data into the geometric buffer.
*/
void GBufferEncode(SurfaceInfo surface, out vec4 color1, out vec4 color2, out vec4 color3, out vec4 color4)
{
    color1 = vec4(surface.Position, surface.AmbientOcclusion);
    color2 = vec4(surface.Normal, 0);
    color3 = vec4(surface.BaseColor, 0);
    color4 = vec4(surface.Metallic, surface.Roughness, surface.Specular, surface.Emission);
}

/*
** Retrieves the surface data from the geometric buffer.
*/
SurfaceInfo GBufferDecode(vec4 color1, vec4 color2, vec4 color3, vec4 color4)
{
	SurfaceInfo surface;
    surface.Position         = color1.xyz;
    surface.Normal           = color2.xyz;
    surface.ViewDir          = normalize(-surface.Position);
    surface.BaseColor        = color3.rgb;
    surface.AmbientOcclusion = color1.a;
    surface.Metallic         = color4.r;
    surface.Roughness        = color4.g;
    surface.Specular         = color4.b;
    surface.Emission         = color4.a;
    surface.DiffuseColor     = SurfaceCalcDiffuse(surface.BaseColor, surface.Metallic);
    surface.SpecularColor    = SurfaceCalcSpecular(surface.BaseColor, surface.Metallic, surface.Specular);
	return surface;
}

// --- Shading model functions ------------------------------------------------

/*
** Evaluates the diffuse BRDF.
*/
vec3 Diffuse(vec3 diffuseColor)
{
    return diffuseColor * RECPI;
}

/*
** Evaluates the specular's normal distribution function (NDF).
*/
float SpecularD(float roughness, float NdotH)
{
    float a = sqr(roughness);
    float a2 = sqr(a);
    float d = ((NdotH * NdotH) * (a2 - 1) + 1);
    
    return a2 / (PI * (d * d));
}

/*
** Evaluates the specular's geometric attenuation term.
*/
float SpecularG(float roughness, float NdotV, float NdotL)
{
    float k = sqr(roughness + 1) * 0.125;
    
    float schlickV = NdotV / (NdotV * (1 - k) + k);
    float schlickL = NdotL / (NdotL * (1 - k) + k);
    
    return schlickV * schlickL;
}

/*
** Evaluates the specular's fresnel term.
*/
vec3 SpecularF(vec3 specularColor, float VdotH)
{
    return specularColor + (saturate(50.0 * specularColor.g) - specularColor) * exp2((-5.55473 * VdotH - 6.98316) * VdotH);
}

/*
** Evaluates the specular BRDF.
*/
vec3 Specular(vec3 specularColor, float NdotL, float NdotH, float NdotV, float VdotH, float roughness)
{
    float SD = SpecularD(roughness, NdotH);
    float SG = SpecularG(roughness, NdotV, NdotL);
    vec3  SF = SpecularF(specularColor, VdotH);
	
	return (SD * SG * SF) / (4 * max(EPSILON, NdotL * NdotV));
}

// --- Lighting model functions -----------------------------------------------

/*
** Computes the light's falloff at the given distance with the desired maximum radius.
*/
float LightFalloff(float distance, float radius)
{
	float d2 = sqr(distance);
    float dr = distance / radius;
    float n = saturate(1 - dr * dr * dr * dr);
    return (n * n) / (d2 + 1);
}

/*
** Computes the light contribution of the ambient hemisphere.
*/
vec3 AmbientHemisphere(AmbientLightInfo ambient, SurfaceInfo surface)
{
	float d = dot(surface.Normal, ambient.Up) * 0.5 + 0.5;
    return surface.BaseColor * mix(ambient.Ground, ambient.Sky, d);
}

/*
** Computes the light contribution of a single directional light.
*/
vec3 LightDirectional(LightInfo light, SurfaceInfo surface)
{
    vec3 N = surface.Normal;
    vec3 V = surface.ViewDir;
    vec3 L = light.Normal;
    vec3 H = normalize(V + L);

    float NdotL = max(0, dot(N, L));
    float NdotV = max(0, dot(N, V));
    float NdotH = max(0, dot(N, H));
    float VdotH = max(0, dot(V, H));
    
    vec3 D = Diffuse(surface.DiffuseColor);
    vec3 S = Specular(surface.SpecularColor, NdotL, NdotH, NdotV, VdotH, surface.Roughness);

    return light.Irradiance * NdotL * (D + S);
}

/*
** Computes the light contribution of a single point light.
*/
vec3 LightPoint(LightInfo light, SurfaceInfo surface)
{
    vec3 N = surface.Normal;
    vec3 V = surface.ViewDir;
	vec3 R = reflect(-V, N);
    vec3 L = normalize(light.Position.xyz - surface.Position);
	vec3 SL = L;

	if (light.Parameters.y > 0)
		SL = normalize(RaySphereClosest(surface.Position, R, vec4(light.Position.xyz, light.Parameters.y)) - surface.Position);
	
	vec3 H = normalize(V + SL);
	
    float d = max(0, distance(surface.Position, light.Position.xyz) - light.Parameters.y);
    float f = LightFalloff(d, light.Parameters.x);

    float NdotL  = max(0, dot(N, L));
	float NdotSL = max(0, dot(N, SL));
    float NdotV  = max(0, dot(N, V));
    float NdotH  = max(0, dot(N, H));
    float VdotH  = max(0, dot(V, H));
    
    vec3 D = Diffuse(surface.DiffuseColor);
    vec3 S = Specular(surface.SpecularColor, NdotSL, NdotH, NdotV, VdotH, surface.Roughness);

    return f * light.Irradiance * NdotL * (D + S);
}

/*
** Computes the light contribution of a single spot light.
*/
vec3 LightSpot(LightInfo light, SurfaceInfo surface)
{
    vec3 N = surface.Normal;
    vec3 V = surface.ViewDir;
    vec3 L = normalize(light.Position.xyz - surface.Position);
    vec3 H = normalize(V + L);
	
    float d = max(0, distance(surface.Position, light.Position.xyz) - light.Parameters.y);
    float f = LightFalloff(d, light.Parameters.x);

	float angle = degrees(acos(-dot(light.Normal, L)));
	f *= 1 - smoothstep(light.Parameters.y, light.Parameters.z, angle);
	
    float NdotL = max(0, dot(N, L));
    float NdotV = max(0, dot(N, V));
    float NdotH = max(0, dot(N, H));
    float VdotH = max(0, dot(V, H));
    
    vec3 D = Diffuse(surface.DiffuseColor);
    vec3 S = Specular(surface.SpecularColor, NdotL, NdotH, NdotV, VdotH, surface.Roughness);

    return f * light.Irradiance * NdotL * (D + S);
}

vec3 Light(LightInfo light, SurfaceInfo surface)
{
    int lightType = int(light.Position.w);
    
	switch (lightType)
	{
		case LIGHT_DIRECTIONAL: return LightDirectional(light, surface);
		case LIGHT_POINT: return LightPoint(light, surface);
		case LIGHT_SPOT: return LightSpot(light, surface);
	}
}

#endif /* COMMON_INC */
