#version 420

// ----------------------------------------------------------------------------
// Includes
// ----------------------------------------------------------------------------

#include <Common.inc>

// ----------------------------------------------------------------------------
// Constants
// ----------------------------------------------------------------------------

const int kNumCoarseSteps = 50;
const int kNumFineSteps = 10;

const float kInitialStep = 0.1;
const float kCoarseStep = 0.1;

const vec2 kZero2 = vec2(0, 0);
const vec2 kOne2 = vec2(1, 1);

// ----------------------------------------------------------------------------
// Attributes
// ----------------------------------------------------------------------------

in vec3 vPosition;
in vec2 vTexCoord;

out vec4 fFragColor;

// ----------------------------------------------------------------------------
// Uniforms
// ----------------------------------------------------------------------------

layout (binding = 0) uniform sampler2D uInputMap;
layout (binding = 1) uniform sampler2D uColorMap1;
layout (binding = 2) uniform sampler2D uColorMap2;
layout (binding = 3) uniform sampler2D uColorMap3;

// ----------------------------------------------------------------------------
// Functions
// ----------------------------------------------------------------------------

vec2 ImageSpaceCoords(vec3 coords)
{
	vec4 projCoord = uScene.Proj * vec4(coords, 1);
	projCoord.xy /= projCoord.w;
	return projCoord.xy * 0.5 + 0.5;
}

vec4 RaycastBinarySearch(vec3 position, vec3 direction)
{
	float depth;
	float difference;
	
	for (int i = 0; i < kNumFineSteps; i++)
	{
		direction *= 0.5;
		position -= direction;
		
		vec2 texCoords = ImageSpaceCoords(position);
		
		depth = texture(uColorMap1, texCoords).z;
		
		if (depth < position.z)
			position += direction;
	}
	
	vec2 texCoords = ImageSpaceCoords(position);	
	return vec4(texCoords, depth, 1);
}

vec4 Raycast(vec3 position, vec3 direction)
{
	vec3 oldPosition;
	
	direction *= kCoarseStep;
	
	for (int i = 0; i < kNumCoarseSteps; i++)
	{
		oldPosition = position;
		position += direction;
		
		vec2 texCoords = ImageSpaceCoords(position);
	
		if (texCoords.x < 0 || texCoords.y < 0 || texCoords.x >= 1 || texCoords.y > 1)
			break;
		
		float depth = texture(uColorMap1, texCoords).z;
		
		float delta = depth - position.z;
		
		//if (abs(delta) < kCoarseStep / 2)
		if (delta > 0)
			return RaycastBinarySearch(position, direction);
			//return vec4(texCoords, depth, 1);
	}
	
	return vec4(0, 0, 0, 0);
}

void main()
{
	// Obtain current point and normal.
	vec3 point = texture(uColorMap1, vTexCoord).xyz;
	vec3 normal = normalize(texture(uColorMap2, vTexCoord).xyz);
	vec3 viewDir = normalize(-point);
	
	// Build ray parameters.
	vec3 direction = normalize(reflect(-viewDir, normal));
	vec3 position  = normal * kInitialStep + point;
		
	// Compute intersection.
	vec4 intersection = Raycast(position, direction);
	
	// Compute final color.
	vec3 reflectAlbedo = texture(uInputMap, intersection.xy).xyz;
	vec3 reflectNormal = texture(uColorMap2, intersection.xy).xyz;
	
	float DdotRN = max(0, -dot(direction, reflectNormal));
	float NdotV  = max(0, dot(viewDir, normal));
	
	fFragColor.rgb = reflectAlbedo * (pow(1 - NdotV * NdotV, 4) * pow(1 - DdotRN, 4) * intersection.w * 0.4);
	fFragColor.rgb += texture(uInputMap, vTexCoord).rgb;
	
	fFragColor.a = 1.0;
}
