#version 420

// ----------------------------------------------------------------------------
// Includes
// ----------------------------------------------------------------------------

#include <Common.inc>

// ----------------------------------------------------------------------------
// Structures
// ----------------------------------------------------------------------------

struct MaterialInfo
{
    vec3 BaseColor;
	float Specular;
	float Roughness;
	float Metallic;
	
	#ifdef SHADER_COLOR_MAP
    sampler2D ColorMap;
	#endif
	
	#ifdef SHADER_NORMAL_MAP
    sampler2D NormalMap;
	#endif
	
	#ifdef SHADER_BRDF_MAP
    sampler2D BRDFMap;
	#endif
};

// ----------------------------------------------------------------------------
// Uniforms
// ----------------------------------------------------------------------------

uniform MaterialInfo uMaterial;

// ----------------------------------------------------------------------------
// Attributes
// ----------------------------------------------------------------------------

// --- Inputs -----------------------------------------------------------------

in vec3 vPosition;
in vec3 vNormal;

#ifdef SHADER_NORMAL_MAP
in vec3 vTangent;
in vec3 vBitangent;
#endif

#if defined(SHADER_COLOR_MAP) || defined(SHADER_NORMAL_MAP) || defined(SHADER_BRDF_MAP)
in vec2 vTexCoord;
#endif

// --- Outputs ----------------------------------------------------------------

#ifndef OPTION_DEFERRED
layout(location = 0) out vec4 fFragColor;
#else
layout(location = 0) out vec4 fColor1;
layout(location = 1) out vec4 fColor2;
layout(location = 2) out vec4 fColor3;
layout(location = 3) out vec4 fColor4;
#endif

// ----------------------------------------------------------------------------
// Functions
// ----------------------------------------------------------------------------

void main()
{
    // Fetch textures - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    
    #ifdef SHADER_COLOR_MAP
        vec4 colColorMap = GammaToLinear(texture(uMaterial.ColorMap, vTexCoord));
    #endif
    
    #ifdef SHADER_NORMAL_MAP
        mat3 TBN = mat3(normalize(vTangent), normalize(vBitangent), normalize(vNormal));
        vec4 colNormalMap = texture(uMaterial.NormalMap, vTexCoord);
    #endif
	
	#ifdef SHADER_BRDF_MAP
        vec4 colBRDFMap = texture(uMaterial.BRDFMap, vTexCoord);
    #endif
    
    // Cutout test  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    
    #if defined(SHADER_COLOR_MAP) && defined(SHADER_ALPHA_AS_CUTOUT)
        if (colOpacityMap.r < 0.5)
            discard;
    #endif
    
    // Setup surface  - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    
    // Get surface parameters.
    SurfaceInfo surface;
    surface.Position  = vPosition;
	surface.ViewDir = normalize(-vPosition);
	surface.BaseColor = uMaterial.BaseColor;
	surface.AmbientOcclusion = 1;
	
	#ifdef SHADER_COLOR_MAP
		surface.BaseColor *= colColorMap.rgb;
	#endif
    
    #ifdef SHADER_NORMAL_MAP
		surface.Normal = normalize(TBN * (colNormalMap.rgb * 2 - 1));
	#else
		surface.Normal = normalize(vNormal);
    #endif
	
	#ifdef SHADER_BRDF_MAP
		surface.Roughness = colBRDFMap.r;
		surface.Metallic  = colBRDFMap.g;
		surface.Specular  = colBRDFMap.b;
		surface.Emission  = colBRDFMap.a;
	#else
		surface.Roughness = uMaterial.Roughness;
		surface.Metallic  = uMaterial.Metallic;
		surface.Specular  = uMaterial.Specular;
		surface.Emission  = 0;
    #endif
    
    // Output - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    
	#ifdef OPTION_DEFERRED
		GBufferEncode(surface, fColor1, fColor2, fColor3, fColor4);
	#else
		surface.DiffuseColor  = SurfaceCalcDiffuse(surface.BaseColor, surface.Metallic);
		surface.SpecularColor = SurfaceCalcSpecular(surface.BaseColor, surface.Metallic, surface.Specular);
	
		#ifdef OPTION_AMBIENT_PASS
    
			fFragColor.rgb = AmbientHemisphere(uAmbientLight, surface);
			fFragColor.rgb *= surface.AmbientOcclusion;
			fFragColor.rgb = mix(fFragColor.rgb, surface.BaseColor, surface.Emission);
			
		#else
		
			fFragColor.rgb = vec3(0);
			
			for (int i = 0; i < uNumLights; i++)
				fFragColor.rgb += Light(uLight[i], surface);
		
			fFragColor.rgb *= (1 - surface.Emission);
			
		#endif
		
		fFragColor.rgb = LinearToGamma(fFragColor.rgb);
		fFragColor.a = 1.0f;
	#endif
}